from nlu.pipe.pipe_components import SparkNLUComponent


class EmbeddingsChunker(SparkNLUComponent):

    def __init__(self, annotator_class='chunk_embedder', language='en', component_type='embeddings_chunk', get_default = True, nlp_ref='', model=None, nlu_ref='',lang='en',loaded_from_pretrained_pipe=False):
        if model != None : self.model = model
        else : 
            if annotator_class == 'chunk_embedder' :
                from nlu import ChunkEmbedder
                if get_default : self.model =  ChunkEmbedder.get_default_model()
                else : self.model =  ChunkEmbedder.get_default_model()  # there are no pretrained chunkers, only default 1
        SparkNLUComponent.__init__(self, annotator_class, component_type, nlu_ref,nlp_ref)



import nlu
from nlu.discovery import Discoverer
from nlu.pipe.utils.storage_ref_utils import StorageRefUtils
from typing import List, Tuple, Optional, Dict, Union
import streamlit as st
from nlu.utils.modelhub.modelhub_utils import ModelHubUtils
import numpy as np
import pandas as pd
from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
from nlu.pipe.viz.streamlit_viz.gen_streamlit_code import get_code_for_viz
from nlu.pipe.viz.streamlit_viz.styles import _set_block_container_style
import random
from nlu.pipe.viz.streamlit_viz.streamlit_viz_tracker import StreamlitVizTracker
class WordSimilarityStreamlitBlock():
    @staticmethod
    def display_word_similarity(
            pipe, #nlu pipe
            default_texts: Tuple[str, str] = ("Donald Trump likes to party!", "Angela Merkel likes to party!"),
            threshold: float = 0.5,
            title: Optional[str] = "Embeddings Similarity Matrix &  Visualizations  ",
            sub_tile :Optional[str]="Visualize `word-wise similarity matrix` and calculate `similarity scores` for `2 texts` and every `word embedding` loaded",
            write_raw_pandas : bool = False,
            display_embed_information:bool = True,
            similarity_matrix = True,
            show_algo_select : bool = True,
            dist_metrics:List[str]  =('cosine'),
            set_wide_layout_CSS:bool=True,
            generate_code_sample:bool = False,
            key:str = "NLU_streamlit",
            num_cols:int=2,
            display_scalar_similarities : bool = False ,
            display_similarity_summary:bool = False,
            model_select_position:str = 'side' , # main or side
            show_infos:bool = True,
            show_logo:bool = True,
    ):

        """We visualize the following cases :
        1. Simmilarity between 2 words - > sim (word_emb1, word_emb2)
        2. Simmilarity between 2 sentences -> let weTW stand word word_emb of token T and sentence S
            2.1. Raw token level with merged embeddings -> sim([we11,we21,weT1], [we12,we22,weT2])
            2.2  Autogenerate sentemb, basically does 2.1 in the Spark NLP backend
            2.3 Already using sentence_embedder model -> sim(se1,se2)
        3. Simmilarity between token and sentence -> sim([we11,w21,wT1], se2)
        4. Mirrored 3
         """
        # https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics.pairwise
        StreamlitVizTracker.footer_displayed=False
        try :
            import plotly.express as px
            from sklearn.metrics.pairwise import distance_metrics
        except :st.error("You need the sklearn and plotly package in your Python environment installed for similarity visualizations. Run <pip install sklearn plotly>")
        if set_wide_layout_CSS : _set_block_container_style()
        if title:st.header(title)
        if show_logo :StreamlitVizTracker.show_logo()
        if sub_tile : st.subheader(sub_tile)

        StreamlitVizTracker.loaded_word_embeding_pipes = []
        dist_metric_algos =distance_metrics()
        dist_algos = list(dist_metric_algos.keys())
        # TODO NORMALIZE DISTANCES TO [0,1] for non cosine
        if 'haversine'   in dist_algos    : dist_algos.remove('haversine') # not applicable in >2D
        if 'precomputed' in dist_algos  : dist_algos.remove('precomputed') # Not a dist
        cols = st.beta_columns(2)
        text1 = cols[0].text_input("Text or word1",default_texts[0],key = key)
        text2 = cols[1].text_input("Text or word2",default_texts[1], key=key) if len(default_texts) >1  else cols[1].text_input("Text or word2",'Please enter second string',key = key)
        # exp = st.sidebar.beta_expander("Select additional Embedding Models and distance metric to compare ")
        e_coms = StreamlitUtilsOS.find_all_embed_components(pipe)
        embed_algos_to_load = []
        embed_pipes = [pipe]
        dist_algo_selection = dist_metrics
        if show_algo_select :
            # emb_components_usable = Discoverer.get_components('embed')
            emb_components_usable = [e for e in Discoverer.get_components('embed',True, include_aliases=True) if 'chunk' not in e and 'sentence' not in e]
            loaded_embed_nlu_refs = []
            loaded_storage_refs = []
            loaded_embed_nlu_refs = list(set(loaded_embed_nlu_refs))

            for c in e_coms :
                if not  hasattr(c.info,'nlu_ref'): continue
                r = c.info.nlu_ref
                if 'en.' not in r and 'embed.' not  in r and 'ner' not in r : loaded_embed_nlu_refs.append('en.embed.' + r)
                elif 'en.'  in r and 'embed.' not  in r  and 'ner' not in r:
                    r = r.split('en.')[0]
                    loaded_embed_nlu_refs.append('en.embed.' + r)
                else :
                    loaded_embed_nlu_refs.append(StorageRefUtils.extract_storage_ref(c))
                loaded_storage_refs.append(StorageRefUtils.extract_storage_ref(c))
            for p in StreamlitVizTracker.loaded_word_embeding_pipes :
                if p != pipe : loaded_embed_nlu_refs.append(p.nlu_ref)
            for l in loaded_embed_nlu_refs:
                if l not in emb_components_usable : emb_components_usable.append(l)
            # embed_algo_selection = exp.multiselect("Click to pick additional Embedding Algorithm",options=emb_components_usable,default=loaded_embed_nlu_refs,key = key)
            # dist_algo_selection = exp.multiselect("Click to pick additional Distance Metric", options=dist_algos, default=dist_metrics, key = key)
            emb_components_usable.sort()
            loaded_embed_nlu_refs.sort()
            dist_algos.sort()
            # dist_metrics.sort()
            if model_select_position =='side':
                embed_algo_selection   = st.sidebar.multiselect("Pick additional Word Embeddings for the Similarity Matrix",options=emb_components_usable,default=loaded_embed_nlu_refs,key = key)
                dist_algo_selection =  st.sidebar.multiselect("Pick additional Similarity Metrics ", options=dist_algos, default=dist_metrics, key = key)
            else :
                exp = st.beta_expander("Pick additional Word Embeddings and Similarity Metrics")
                embed_algo_selection   = exp.multiselect("Pick additional Word Embeddings for the Similarity Matrix",options=emb_components_usable,default=loaded_embed_nlu_refs,key = key)
                dist_algo_selection =  exp.multiselect("Pick additional Similarity Metrics ", options=dist_algos, default=dist_metrics, key = key)
            embed_algos_to_load = list(set(embed_algo_selection) - set(loaded_embed_nlu_refs))

        for embedder in embed_algos_to_load:embed_pipes.append(nlu.load(embedder))

        if generate_code_sample:st.code(get_code_for_viz('SIMILARITY',[StreamlitUtilsOS.extract_name(p) for p  in embed_pipes],default_texts))

        StreamlitVizTracker.loaded_word_embeding_pipes+=embed_pipes
        similarity_metrics = {}
        embed_vector_info = {}
        cols_full = True
        col_index=0
        # for p in embed_pipes :
        for p in StreamlitVizTracker.loaded_word_embeding_pipes :
            data1 = p.predict(text1,output_level='token').dropna()
            data2 = p.predict(text2,output_level='token').dropna()
            e_coms = StreamlitUtilsOS.find_all_embed_components(p)
            modelhub_links = [ModelHubUtils.get_url_by_nlu_refrence(c.info.nlu_ref) if hasattr(c.info,'nlu_ref') else ModelHubUtils.get_url_by_nlu_refrence('') for c in e_coms]
            e_cols = StreamlitUtilsOS.get_embed_cols(p)
            for num_emb,e_col in enumerate(e_cols):
                if col_index == num_cols-1 :cols_full=True
                if cols_full :
                    cols = st.beta_columns(num_cols)
                    col_index = 0
                    cols_full = False
                else:col_index+=1
                tok1 = data1['token']
                tok2 = data2['token']
                emb1 = data1[e_col]
                emb2 = data2[e_col]
                embed_mat1 = np.array([x for x in emb1])
                embed_mat2 = np.array([x for x in emb2])
                # e_name = e_col.split('word_embedding_')[-1]
                e_name = e_coms[num_emb].info.nlu_ref if hasattr(e_coms[num_emb].info,'nlu_ref') else e_col.split('word_embedding_')[-1] if 'en.' in e_col else e_col
                e_name = e_name.split('embed.')[-1] if 'en.' in e_name else e_name
                if 'ner' in e_name : e_name = loaded_storage_refs[num_emb]

                embed_vector_info[e_name]= {"Vector Dimension ":embed_mat1.shape[1],
                                            "Num Vectors":embed_mat1.shape[0] + embed_mat1.shape[0],
                                            "NLU_reference":e_coms[num_emb].info.nlu_ref if hasattr(e_coms[num_emb].info,'nlu_ref') else ' ',
                                            "Spark_NLP_reference":ModelHubUtils.NLU_ref_to_NLP_ref(e_coms[num_emb].info.nlu_ref if hasattr(e_coms[num_emb].info,'nlu_ref') else ' '),
                                            "Storage Reference":loaded_storage_refs[num_emb],
                                            'Modelhub info': modelhub_links[num_emb]}
                for dist_algo in dist_algo_selection:
                    # scalar_similarities[e_col][dist_algo]={}
                    sim_score = dist_metric_algos[dist_algo](embed_mat1,embed_mat2)
                    sim_score = pd.DataFrame(sim_score)
                    sim_score.index   = tok1.values
                    sim_score.columns = tok2.values
                    sim_score.columns = StreamlitVizTracker.pad_duplicate_tokens(list(sim_score.columns))
                    sim_score.index   = StreamlitVizTracker.pad_duplicate_tokens(list(sim_score.index))
                    if write_raw_pandas :st.write(sim_score,key = key)
                    if sim_score.shape == (1,1) :
                        sim_score = sim_score.iloc[0][0]
                        sim_score = round(sim_score,2)
                        if sim_score > threshold:
                            st.success(sim_score)
                            st.success(f'Scalar Similarity={sim_score} for distance metric={dist_algo}')
                            st.error('No similarity matrix for only 2 tokens. Try entering at least 1 sentences in a field')
                        else:
                            st.error(f'Scalar Similarity={sim_score} for distance metric={dist_algo}')
                    else :
                        ploty_avaiable = True
                        # for tok emb, sum rows and norm by rows, then sum cols and norm by cols to generate a scalar from matrix
                        scalar_sim_score  = np.sum((np.sum(sim_score,axis=0) / sim_score.shape[0])) / sim_score.shape[1]
                        scalar_sim_score = round(scalar_sim_score,2)

                        if display_scalar_similarities:
                            if scalar_sim_score > threshold:st.success(f'Scalar Similarity :{scalar_sim_score} for distance metric={dist_algo}')
                            else: st.error(f'Scalar Similarity :{scalar_sim_score} for embedder={e_col} distance metric={dist_algo}')
                        if similarity_matrix:
                            if ploty_avaiable :
                                fig = px.imshow(sim_score, labels=dict(color="similarity"))#, title=f'Simmilarity Matrix for embedding_model={e_name} distance metric={dist_algo}')
                                # st.write(fig,key =key)
                                similarity_metrics[f'{e_name}_{dist_algo}_similarity']={
                                    'scalar_similarity' : scalar_sim_score,
                                    'dist_metric' : dist_algo,
                                    'embedding_model': e_name,
                                    'modelhub_info' : modelhub_links[num_emb],
                                }
                                subh = f"""Embedding-Model=`{e_name}`, Similarity-Score=`{scalar_sim_score}`,  distance metric=`{dist_algo}`"""
                                cols[col_index].markdown(subh)
                                cols[col_index].write(fig, key=key)
                            else : pass # todo fallback plots

        if display_similarity_summary:
            exp = st.beta_expander("Similarity summary")
            exp.write(similarity_metrics)
        if display_embed_information:
            exp = st.beta_expander("Embedding vector information")
            exp.write(embed_vector_info)
        if show_infos :
            # VizUtilsStreamlitOS.display_infos()
            StreamlitVizTracker.display_model_info(pipe.nlu_ref, pipes = [pipe])
            StreamlitVizTracker.display_footer()



from sparknlp.annotator import *


class NERDLCRF:
    @staticmethod
    def get_default_model():
        return NerCrfModel.pretrained(name="ner_dl_bert") \
            .setInputCols(["sentence", "token", "pos", "word_embeddings"]) \
            .setOutputCol("ner")

    @staticmethod
    def get_pretrained_model(name, language):
        return NerCrfModel.pretrained(name,language) \
            .setInputCols(["sentence", "token", "pos", "word_embeddings"]) \
            .setOutputCol("ner")

    @staticmethod
    def get_default_trainable_model():
        return NerCrfApproach() \
            .setInputCols(["sentence", "token", "pos"]) \
            .setLabelColumn("label") \
            .setOutputCol("ner") \
            .setMinEpochs(1) \
            .setMaxEpochs(20) \
            .setLossEps(1e-3) \
            .setDicts(["ner-corpus/dict.txt"]) \
            .setL2(1) \
            .setC0(1250000) \
            .setRandomSeed(0) \
            .setVerbose(2)


from sparknlp.annotator import *
import inspect
import logging
logger = logging.getLogger('nlu')
from nlu.pipe.utils.component_utils import ComponentUtils
import nlu
import requests
class ModelHubUtils():
    modelhub_json_url = 'https://nlp.johnsnowlabs.com/models.json'
    data = requests.get(modelhub_json_url).json()
    """Pipe Level logic oprations and utils"""
    @staticmethod
    def NLU_ref_to_NLP_ref(nlu_ref: str,lang: str =None) -> str:
        """Resolve a Spark NLU reference to a NLP reference.

        Args :
        NLU_ref : which nlu model's nlp refrence to return.
        lang : what language is the model in.
        """
        nlu_namespaces_to_check = [nlu.Spellbook.pretrained_pipe_references, nlu.Spellbook.pretrained_models_references, nlu.Spellbook.pretrained_healthcare_model_references, nlu.Spellbook.licensed_storage_ref_2_nlu_ref , nlu.Spellbook.storage_ref_2_nlu_ref]#component_alias_references  ]
        for dict_ in nlu_namespaces_to_check:
            if lang:
                if lang in dict_.keys():
                    for reference in dict_[lang]:
                        if reference ==nlu_ref:
                            return  dict_[lang][reference]
            else :
                for dict_ in nlu_namespaces_to_check:
                    for lang in dict_:
                        for reference in dict_[lang]:
                            if reference ==nlu_ref:
                                return  dict_[lang][reference]



    def get_url_by_nlu_refrence(nlu_refrence: str ) -> str:
        """Rsolves  a  URL for an NLU refrence.

        Args :
            nlu_refrence : Which nlu refrence's url to return.

        """

        # getting spark refrence for given nlu refrence
        nlp_refrence= ModelHubUtils.NLU_ref_to_NLP_ref(nlu_refrence)
        if nlp_refrence == None :
            print(nlp_refrence," 			", nlu_refrence)
        for model in ModelHubUtils.data :
            if (model['language'] in nlu_refrence.split(".") or model['language'] in nlp_refrence.split('_')) and model['name'] == nlp_refrence:

                return f"https://nlp.johnsnowlabs.com/{model['url']}"


    def return_json_entry(nlp_refrence:str,language:str) -> dict:
        """Resolves a Json entry for an nlp_refrence.

        Args:
            nlp_refrence: What nlp_refrence to resolve.
            language : Which language the model is in.
        """

        for model in ModelHubUtils.data :
            if model['language']== language and model["name"] == nlp_refrence:
                return model



import nlu
from nlu.discovery import Discoverer
from nlu.pipe.utils.storage_ref_utils import StorageRefUtils
from typing import List, Tuple, Optional, Dict, Union
import streamlit as st
from nlu.utils.modelhub.modelhub_utils import ModelHubUtils
import numpy as np
import pandas as pd
from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
from nlu.pipe.viz.streamlit_viz.gen_streamlit_code import get_code_for_viz
from nlu.pipe.viz.streamlit_viz.styles import _set_block_container_style
import random
from nlu.pipe.viz.streamlit_viz.streamlit_viz_tracker import StreamlitVizTracker
class NERStreamlitBlock():
    @staticmethod
    def visualize_ner(
            pipe, # Nlu pipe
            text:str,
            ner_tags: Optional[List[str]] = None,
            show_label_select: bool = True,
            show_table: bool = False,
            title: Optional[str] = "Named Entities",
            sub_title: Optional[str] = "Recognize various `Named Entities (NER)` in text entered and filter them. You can select from over `100 languages` in the dropdown.",
            colors: Dict[str, str] = {},
            show_color_selector: bool = False,
            set_wide_layout_CSS:bool=True,
            generate_code_sample:bool = False,
            key = "NLU_streamlit",
            model_select_position:str = 'side',
            show_model_select : bool = True,
            show_text_input:bool = True,
            show_infos:bool = True,
            show_logo:bool = True,

    ):
        StreamlitVizTracker.footer_displayed=False
        if set_wide_layout_CSS : _set_block_container_style()
        if show_logo :StreamlitVizTracker.show_logo()
        if show_model_select :
            model_selection = Discoverer.get_components('ner',include_pipes=True)
            model_selection.sort()
            if model_select_position == 'side':ner_model_2_viz = st.sidebar.selectbox("Select a NER model",model_selection,index=model_selection.index(pipe.nlu_ref.split(' ')[0]))
            else : ner_model_2_viz = st.selectbox("Select a NER model",model_selection,index=model_selection.index(pipe.nlu_ref.split(' ')[0]))
            pipe = pipe if pipe.nlu_ref == ner_model_2_viz else StreamlitUtilsOS.get_pipe(ner_model_2_viz)
        if title: st.header(title)
        if show_text_input : text = st.text_area("Enter text you want to visualize NER classes for below", text, key=key)
        if sub_title : st.subheader(sub_title)
        if generate_code_sample: st.code(get_code_for_viz('NER',StreamlitUtilsOS.extract_name(pipe),text))
        if ner_tags is None: ner_tags = StreamlitUtilsOS.get_NER_tags_in_pipe(pipe)

        if not show_color_selector :
            if show_label_select:
                exp = st.beta_expander("Select entity labels to highlight")
                label_select = exp.multiselect(
                    "These labels are predicted by the NER model. Select which ones you want to display",
                    options=ner_tags,default=list(ner_tags))
            else : label_select = ner_tags
            pipe.viz(text,write_to_streamlit=True, viz_type='ner',labels_to_viz=label_select,viz_colors=colors, streamlit_key=key)
        else : # TODO WIP color select
            cols = st.beta_columns(3)
            exp = cols[0].beta_expander("Select entity labels to display")
            color = st.color_picker('Pick A Color', '#00f900',key = key)
            color = cols[2].color_picker('Pick A Color for a specific entity label', '#00f900',key = key)
            tag2color = cols[1].selectbox('Pick a ner tag to color', ner_tags,key = key)
            colors[tag2color]=color
        if show_table : st.write(pipe.predict(text, output_level='chunk'),key = key)

        if show_infos :
            # VizUtilsStreamlitOS.display_infos()
            StreamlitVizTracker.display_model_info(pipe.nlu_ref, pipes = [pipe])
            StreamlitVizTracker.display_footer()




from sparknlp.annotator import *

class Roberta:
    @staticmethod
    def get_default_model():
        return RoBertaEmbeddings.pretrained() \
        .setInputCols("sentence", "token") \
        .setOutputCol("roberta")

    @staticmethod
    def get_pretrained_model(name, language):
        return RoBertaEmbeddings.pretrained(name, language) \
            .setInputCols("sentence", "token") \
            .setOutputCol("roberta")






"""Deduct name for annotator"""
"""
Resolve Annotator Classes in the Pipeline to Extractor Configs and Methods.
Each Spark NLP Annotator Class is mapped to at least one

Every Annotator should have 2 configs. Some might offor multuple configs/method pairs, based on model/NLP reference.
- default/minimalistic -> Just the results of the annotations, no confidences or extra metadata
- with meta            -> A config that leverages white/black list and gets the most relevant metadata
- with positions       -> With Begins/Ends


"""
from sparknlp.annotator import *
from sparknlp.base import *
name_deductable_OS = [
    NerConverter,
    BertEmbeddings,
    AlbertEmbeddings,
    XlnetEmbeddings ,
    WordEmbeddingsModel ,
    ElmoEmbeddings ,
    BertSentenceEmbeddings,
    UniversalSentenceEncoder,
    SentenceEmbeddings,
    ContextSpellCheckerModel ,
    SymmetricDeleteModel ,
    NorvigSweetingModel ,
    NerDLModel ,
    NerCrfModel,
    LanguageDetectorDL ,
    SentimentDLModel ,
    SentimentDetectorModel ,
    ViveknSentimentModel ,
    MultiClassifierDLModel,
    ClassifierDLModel ,
    ChunkEmbeddings ,
    TextMatcherModel,
    RegexMatcherModel,
    DateMatcher,
    MultiDateMatcher,
    T5Transformer,
    MarianTransformer,
    WordSegmenterModel,


    DistilBertEmbeddings,
    RoBertaEmbeddings,
    XlmRoBertaEmbeddings,


    # approaches
    ViveknSentimentApproach    ,
    SentimentDLApproach        ,
    ClassifierDLApproach       ,
    MultiClassifierDLApproach  ,
    NerDLApproach              ,
    PerceptronApproach         ,
    Doc2Chunk,
]


always_name_deductable_OS = [
    BertEmbeddings,
    AlbertEmbeddings,
    XlnetEmbeddings ,
    WordEmbeddingsModel ,
    ElmoEmbeddings ,
    BertSentenceEmbeddings,
    UniversalSentenceEncoder,
    SentenceEmbeddings,
    MultiClassifierDLModel,
    ClassifierDLModel ,
    ChunkEmbeddings ,
    TextMatcherModel,
    RegexMatcherModel,
    DateMatcher,
    MultiDateMatcher,
    # T5Transformer,
    # MarianTransformer,
    # WordSegmenterModel,


    DistilBertEmbeddings,
    RoBertaEmbeddings,
    XlmRoBertaEmbeddings,



]















from nlu.pipe.pipe_components import SparkNLUComponent

class SpellChecker(SparkNLUComponent):
    def __init__(self, annotator_class='context_spell', language ='en', component_type='spell_checker', get_default=True, model = None, nlp_ref='', dataset='', nlu_ref ='', is_licensed=False):
        if annotator_class == 'context' or annotator_class == 'norvig' or annotator_class == 'symmetric':
            annotator_class = annotator_class + '_spell'
        if dataset != '':annotator_class = dataset + '_spell'
        if 'spellcheck_dl' in nlp_ref : annotator_class ='context_spell'
        if 'spell.med' in nlu_ref : annotator_class ='context'
        if 'spell.clinical' in nlu_ref : annotator_class ='context'
        if '.med' in nlu_ref : annotator_class ='context'

        if model != None : self.model = model
        else :
            if 'context' in annotator_class:
                from nlu import ContextSpellChecker
                if is_licensed : self.model =  ContextSpellChecker.get_pretrained_model(nlp_ref, language,bucket='clinical/models')
                elif get_default : self.model =  ContextSpellChecker.get_default_model()
                else : self.model = ContextSpellChecker.get_pretrained_model(nlp_ref, language)
            elif 'norvig' in annotator_class:
                from nlu import NorvigSpellChecker
                if get_default : self.model =  NorvigSpellChecker.get_default_model()
                else : self.model = NorvigSpellChecker.get_pretrained_model(nlp_ref, language)
            elif 'symmetric' in annotator_class :
                from nlu import SymmetricSpellChecker
                if get_default : self.model = SymmetricSpellChecker.get_default_model()
                else : self.model = SymmetricSpellChecker.get_pretrained_model(nlp_ref, language)

        SparkNLUComponent.__init__(self, annotator_class, component_type)



"""
Resolve Annotator Classes in the Pipeline to Extractor Configs and Methods

Every Annotator should have 2 configs. Some might offor multuple configs/method pairs, based on model/NLP reference.
- default/minimalistic -> Just the results of the annotations, no confidences or extra metadata
- with meta            -> A config that leverages white/black list and gets the most relevant metadata
- with positions       -> With Begins/Ends
- with sentence references -> Reeturn the sentence/chunk no. reference from the metadata.
                                If a document has multi-sentences, this will map a label back to a corrosponding sentence

"""
from nlu.pipe.col_substitution.col_substitution_HC import *
from nlu.pipe.col_substitution.col_substitution_OS import *

from sparknlp_jsl.annotator  import *

HC_anno2substitution_fn = {
    MedicalNerModel : {
        'default': substitute_ner_dl_cols ,
    },
    NerConverterInternal : {
        'default': substitute_ner_internal_converter_cols,
    },
    AssertionDLModel : {
        'default': substitute_assertion_cols,
    },
    AssertionLogRegModel : {
        'default': substitute_assertion_cols,
    },
    SentenceEntityResolverModel : {
        'default': substitute_sentence_resolution_cols,
    },
    ChunkEntityResolverModel : {
        'default': substitute_chunk_resolution_cols,
    },
    ChunkEntityResolverApproach : {
        'default': substitute_chunk_resolution_cols,
    },


    DeIdentificationModel : {
        'default': substitute_de_identification_cols,
    },
    RelationExtractionModel : {
        'default': substitute_relation_cols,
    },

    RelationExtractionDLModel : {
        'default': substitute_relation_cols,
    },
    Chunk2Token : {
        'default': '',# TODO
    },

    ContextualParserApproach : {
        'default' :substitute_context_parser_cols,#

    },
    ContextualParserModel : {
        'default': substitute_context_parser_cols,

    },

    DrugNormalizer : {
        'default': substitute_drug_normalizer_cols
    },

    GenericClassifierModel : {
        'default': substitute_generic_classifier_parser_cols,
    },
    GenericClassifierApproach : {
        'default': substitute_generic_classifier_parser_cols,
    },


    ChunkMergeModel : {
        'default': '',# TODO
    },

    NerDisambiguatorModel : {
        'default': '',# TODO
    },

    RENerChunksFilter : {
        'default': '',# TODO
    },

    NerOverwriter : {
        'default': '',# TODO
    },
    PosologyREModel : {
        'default': substitute_relation_cols,
    }

}










"""
Resolve Annotator Classes in the Pipeline to Extractor Configs and Methods.
Each Spark NLP Annotator Class is mapped to at least one

Every Annotator should have 2 configs. Some might offor multuple configs/method pairs, based on model/NLP reference.
- default/minimalistic -> Just the results of the annotations, no confidences or extra metadata
- with meta            -> A config that leverages white/black list and gets the most relevant metadata
- with positions       -> With Begins/Ends


"""
from sparknlp.annotator import *
from sparknlp.base import *
from nlu.pipe.extractors.extractor_configs_open_source import *

OS_anno2config = {
    NerConverter : {
        'default': default_ner_converter_config ,
        'default_full'  : default_full_config,
    },
    MultiClassifierDLModel : {
        'default': default_multi_classifier_dl_config ,
        'default_full'  : default_full_config,

    },
    PerceptronModel : {
        'default': default_POS_config,
        'default_full'  : default_full_config,
    },
    ClassifierDLModel : {
        'default': default_classifier_dl_config,
        'default_full'  : default_full_config,
    },
    BertEmbeddings : {
        'default': default_word_embedding_config,
        'default_full'  : default_full_config,
    },

    AlbertEmbeddings : {
        'default': default_word_embedding_config,
        'default_full'  : default_full_config,
    },
    XlnetEmbeddings : {
        'default': default_word_embedding_config,
        'default_full'  : default_full_config,
    },
    RoBertaEmbeddings : {
        'default': default_word_embedding_config,
        'default_full'  : default_full_config,
    },
    XlmRoBertaEmbeddings : {
        'default': default_word_embedding_config,
        'default_full'  : default_full_config,
    },
    DistilBertEmbeddings : {
        'default': default_word_embedding_config,
        'default_full'  : default_full_config,
    },
    WordEmbeddingsModel : {
        'default': default_word_embedding_config,
        'default_full'  : default_full_config,
    },
    ElmoEmbeddings : {
        'default': default_word_embedding_config,
        'default_full'  : default_full_config,
    },
    BertSentenceEmbeddings : {
        'default': default_sentence_embedding_config,
        'default_full'  : default_full_config,
    },
    UniversalSentenceEncoder : {
        'default': default_sentence_embedding_config,
        'default_full'  : default_full_config,
    },
    SentenceEmbeddings : {
        'default': default_sentence_embedding_config,
        'default_full'  : default_full_config,
    },
    Tokenizer : {
        'default': default_tokenizer_config,
        'default_full'  : default_full_config,
    },
    TokenizerModel : {
        'default': default_tokenizer_config,
        'default_full'  : default_full_config,
    },
    RegexTokenizer : {
        'default': default_tokenizer_config,
        'default_full'  : default_full_config,
    },
    DocumentAssembler : {
        'default': default_document_config,
        'default_full'  : default_full_config,
    },
    SentenceDetectorDLModel : {
        'default': default_sentence_detector_DL_config,
        'default_full'  : default_full_config,
    },
    SentenceDetector : {
        'default': default_sentence_detector_config,
        'default_full'  : default_full_config,
    },
    ContextSpellCheckerModel : {
        'default': default_spell_context_config,
        'default_full'  : default_full_config,
    },
    SymmetricDeleteModel : {
        'default': default_spell_symmetric_config,
        'default_full'  : default_full_config,
    },
    NorvigSweetingModel : {
        'default': default_spell_norvig_config,
        'default_full'  : default_full_config,
    },
    LemmatizerModel : {
        'default': default_lemma_config,
        'default_full'  : default_full_config,
    },
    Normalizer : {
        'default': default_norm_config,
        'default_full'  : default_full_config,
    },
    NormalizerModel : {
        'default': default_norm_config,
        'default_full'  : default_full_config,
    },
    DocumentNormalizer : {
        'default':default_norm_document_config,
        'default_full'  : default_full_config,
    },
    Stemmer : {
        'default': default_stemm_config,
        'default_full'  : default_full_config,
    },
    NerDLModel : {
        'default': default_NER_config,
        'meta': meta_NER_config,
        'default_full'  : default_full_config,
    },
    NerCrfModel : {
        'default': '',# TODO
        'default_full'  : default_full_config,
    },
    LanguageDetectorDL : {
        'default': default_lang_classifier_config,
        'default_full'  : default_full_config,
    },
    DependencyParserModel : {
        'default': default_dep_untyped_config,
        'default_full'  : default_full_config,
    },
    TypedDependencyParserModel : {
        'default': default_dep_typed_config,
        'default_full'  : default_full_config,
    },
    SentimentDLModel : {
        'default': default_sentiment_dl_config,
        'default_full'  : default_full_config,
    },
    SentimentDetectorModel : {
        'default': default_sentiment_config,
        'default_full'  : default_full_config,
    },
    ViveknSentimentModel : {
        'default': default_sentiment_vivk_config,
        'default_full'  : default_full_config,
    },
    Chunker : {
        'default': default_chunk_config,
        'default_full'  : default_full_config,
    },
    NGramGenerator : {
        'default': default_ngram_config,
        'default_full'  : default_full_config,
    },
    ChunkEmbeddings : {
        'default': default_chunk_embedding_config,
        'default_full'  : default_full_config,
    },
    StopWordsCleaner : {
        'default': default_stopwords_config,
        'default_full'  : default_full_config,
    },
    TextMatcherModel : {
        'default': '',# TODO
        'default_full'  : default_full_config,
    },
    TextMatcher : {
        'default': '',# TODO
        'default_full'  : default_full_config,
    },

    RegexMatcherModel : {
        'default': '',# TODO
        'default_full'  : default_full_config,
    },
    RegexMatcher : {
        'default': '',# TODO
        'default_full'  : default_full_config,
    },
    DateMatcher : {
        'default':'',# TODO
        'default_full'  : default_full_config,
    },

    MultiDateMatcher : {
        'default': '',# TODO
        'default_full'  : default_full_config,
    },

    Doc2Chunk : {
        'default': default_doc2chunk_config,
        'default_full'  : default_full_config,
    },


    T5Transformer : {
        'default': default_T5_config,
        'default_full'  : default_full_config,
    },
    MarianTransformer : {
        'default': default_marian_config,
        'default_full'  : default_full_config,
    },
    YakeModel : {
        'default': default_yake_config,
        'default_full'  : default_full_config,
    },
    WordSegmenterModel : {
        'default': default_word_segmenter_config,
        'default_full'  : default_full_config,
    },



    # approaches
    ViveknSentimentApproach    :{'default':'' , 'default_full'  : default_full_config,},
    SentimentDLApproach        :{'default':default_sentiment_dl_config , 'default_full'  : default_full_config,},
    ClassifierDLApproach        :{'default':default_classifier_dl_config , 'default_full'  : default_full_config,},
    MultiClassifierDLApproach  :{'default':default_multi_classifier_dl_config, 'default_full'  : default_full_config,},
    NerDLApproach              :{'default':default_NER_config , 'default_full'  : default_full_config,},
    PerceptronApproach         :{'default':default_POS_config , 'default_full'  : default_full_config,},


    # PretrainedPipeline : {
    #     'default' : '',
    # }
}










from nlu.pipe.pipe_components import SparkNLUComponent
class Relation(SparkNLUComponent):
    def __init__(self, annotator_class='relation_extractor', lang='en', component_type='relation_extractor', get_default=True, model = None, nlp_ref ='', nlu_ref='', trainable=False, is_licensed=False,loaded_from_pretrained_pipe=False):

        if 're_' in nlp_ref  : annotator_class = 'relation_extractor'
        if 'redl' in nlp_ref : annotator_class = 'relation_extractor_dl'

        if model != None : self.model = model
        else :
            if annotator_class == 'relation_extractor':
                from nlu.components.relation_extractors.relation_extractor.relation_extractor import RelationExtraction
                if trainable : self.model = RelationExtraction.get_default_trainable_model()
                else : self.model = RelationExtraction.get_pretrained_model(nlp_ref, lang, 'clinical/models')

            elif annotator_class == 'relation_extractor_dl':
                from nlu.components.relation_extractors.relation_extractor_dl.relation_extractor_dl import RelationExtractionDL
                # if trainable : self.model = RelationExtractionDL.get_default_trainable_model()
                self.model = RelationExtractionDL.get_pretrained_model(nlp_ref, lang, 'clinical/models')


        SparkNLUComponent.__init__(self, annotator_class, component_type, nlu_ref, nlp_ref, lang,loaded_from_pretrained_pipe , is_licensed)



import logging
from  nlu.pipe.utils import uid_to_storageref as uid2storageref
logger = logging.getLogger('nlu')


"""Storage Ref logic operations and utils"""
class StorageRefUtils():
    @staticmethod
    def has_storage_ref(component):
        """Storage ref is either on the model or nlu component defined """
        return StorageRefUtils.has_component_storage_ref_or_anno_storage_ref(component)


    @staticmethod
    def extract_storage_ref(component, prefer_anno=False):
        """Extract storage ref from either a NLU component or NLP Annotator. First cheks if annotator has storage ref, otherwise check NLU attribute"""
        if StorageRefUtils.has_storage_ref(component):
            if hasattr(component.info,'storage_ref') and not prefer_anno: return StorageRefUtils.fallback_storage_ref_resolutions(component.info.storage_ref)
            if StorageRefUtils.nlp_component_has_storage_ref(component.model) :
                return StorageRefUtils.fallback_storage_ref_resolutions( StorageRefUtils.nlu_extract_storage_ref_nlp_model(component))
        return ''

    @staticmethod
    def fallback_storage_ref_resolutions(storage_ref):
        """
            For every storage ref result, we check if its storage ref is defined as its UID and if a fallback storageref is avaiable.
            If avaiable, alternative is returned, otherwise original
        """
        if storage_ref in uid2storageref.mappings.keys():
            return uid2storageref.mappings[storage_ref]
        else : return storage_ref


    ### SUB HELPERS

    @staticmethod
    def has_component_storage_ref_or_anno_storage_ref(component):
        """Storage ref is either on the model or nlu component defined """
        if StorageRefUtils.nlp_component_has_storage_ref(component.model): return True
        if StorageRefUtils.nlu_component_has_storage_ref(component) : return True


    @staticmethod
    def nlp_component_has_storage_ref(model):
        """Check if a storage ref is defined on the Spark NLP Annotator model"""
        for k, _ in model.extractParamMap().items():
            if k.name == 'storageRef': return True
        return False


    @staticmethod
    def extract_storage_ref_from_component(component):
        """Extract storage ref from a NLU component which embelished a Spark NLP Annotator"""
        if StorageRefUtils.nlu_component_has_storage_ref(component):
            return component.info.storage_ref
        elif StorageRefUtils.nlp_component_has_storage_ref(component):
            return StorageRefUtils.nlp_extract_storage_ref_nlp_model(component)
        else:
            return ''


    @staticmethod
    def nlu_extract_storage_ref_nlp_model(component):
        """Extract storage ref from a NLU component which embelished a Spark NLP Annotator"""
        return component.model.extractParamMap()[component.model.getParam('storageRef')]

    @staticmethod
    def nlu_component_has_storage_ref(component):
        """Check if a storage ref is defined on the Spark NLP Annotator embelished by the NLU Component"""
        if hasattr(component.info, 'storage_ref'): return True
        return False

    @staticmethod
    def nlp_extract_storage_ref_nlp_model(model):
        """Extract storage ref from a NLU component which embelished a Spark NLP Annotator"""
        return model.extractParamMap()[model.model.getParam('storageRef')]





import unittest
from tests.test_utils import get_sample_pdf_with_labels, get_sample_pdf, get_sample_sdf, get_sample_pdf_with_extra_cols, get_sample_pdf_with_no_text_col ,get_sample_spark_dataframe
from nlu import *

class TestNormalize(unittest.TestCase):

    def test_document_normalizer_pipe(self):
        pipe = nlu.load('norm_document', verbose=True )
        data = '<!DOCTYPE html> <html> <head> <title>Example</title> </head> <body> <p>This is an example of a simple HTML page with one paragraph.</p> </body> </html>'
        df = pipe.predict(data,output_level='sentence',drop_irrelevant_cols=False, metadata=True, )
        for c in df.columns: print(df[c])


if __name__ == '__main__':
    unittest.main()




"""Utils for making final output columns generated by pythonify nicer
Col naming schema follows
<type>.<nlu_ref_identifier>.<field>
IF there is only 1 component of <type> in the pipe, the <type> will/can be ommitted.

- we remove all _<field> suffixex
- replace all '@' with '_'
"""

from sparknlp.annotator import *
from nlu.pipe.col_substitution import substitution_map_OS
from nlu.pipe.col_substitution import col_substitution_OS
import logging
logger = logging.getLogger('nlu')

""" NAMING SCHEMAS after pythonify procedure : 
### NAMING RESULT SCHEMA: 


results         = { configs.output_col_prefix+'_results'  : list(map(unpack_result,row))} if configs.get_result else {}
beginnings      = { configs.output_col_prefix+'_beginnings' : list(map(unpack_begin,row))} if configs.get_begin or configs.get_positions else {}
endings         = { configs.output_col_prefix+'_endings'    : next(map(unpack_end,row))} if configs.get_end or configs.get_positions else {}
embeddings      = { configs.output_col_prefix+'_embeddings' : next(map(unpack_embeddings,row))} if configs.get_embeds else {}


### METADATA NAMING SCHEMA

result = dict(zip(list(map(lambda x : 'meta_'+ configs.output_col_prefix + '_' + x, keys_in_metadata)),meta_values_list))




"""
from nlu.pipe.pipe_logic import PipeUtils
class ColSubstitutionUtils():
    """Utils for substituting col names in Pythonify to short and meaningful names.
    Uses custom rename methods for either PySpark or Pandas
    """
    from sparknlp.annotator import MarianTransformer
    cleanable_splits = ['ner_converter','spell','ner_to_chunk_converter','train','classify','ner','med_ner','dl','match','clean','sentiment','embed','embed_sentence','embed_chunk','explain','pos','resolve_chunk','resolve',]
    all_langs        = ['vi','mt','ta','af','cy','et','en','et','bh','am','da','fr','de','it','nb','no','nn','pl','pt','ru','es','af','ar','hy','eu','bn','br','bg','ca','cs','eo','fi','gl','el','ha','he','hi','hu','id','ga','ja','la','lv','mr','fa','ro','sk','sl','so','st','sw','sv','th','tr','uk','yo','zu','zh','xx','ur','ko']
    @staticmethod
    def substitute_col_names(df,anno_2_ex,pipe, stranger_cols=[],drop_debug_cols=True,):
        """
        Some truly irrelevant cols might be dropped, regardless of anno Extractor config
        Some truly irrelevant cols might be dropped, regardless of anno Extractor config
        0. Get list of annotator classes that are duplicates. Check inside the NLU Component Embelishment
        1. Get list of cols derived by component
        2. Substitute list of cols in DF with custom logic
        """
        substitution_fn = 'TODO'

        anno2final_cols = {}# mapping of final col names to annotator class Key=AnnoModel, Value=List of Result cols
        new_cols = {}
        if pipe.has_licensed_components :
            from nlu.pipe.col_substitution import col_substitution_HC
            from nlu.pipe.col_substitution import substitution_map_HC
        deducted_component_names = ColSubstitutionUtils.deduct_component_names(pipe)
        for c in pipe.components :
            cols_to_substitute = ColSubstitutionUtils.get_final_output_cols_of_component(c,df,anno_2_ex)

            if type(c.model) in substitution_map_OS.OS_anno2substitution_fn.keys():
                substitution_fn = substitution_map_OS.OS_anno2substitution_fn[type(c.model)]['default']
            if pipe.has_licensed_components and substitution_fn != 'TODO':
                if type(c.model) in substitution_map_HC.HC_anno2substitution_fn.keys():
                    substitution_fn = substitution_map_HC.HC_anno2substitution_fn[type(c.model)]['default']
            if substitution_fn =='TODO':
                logger.info(f"Could not find substitution function for c={c}, leaving col names untouched")
                old2new_anno_cols = dict(zip(cols_to_substitute,cols_to_substitute))
                anno2final_cols[c.model] = list(old2new_anno_cols.values())
                new_cols.update(old2new_anno_cols)
                continue
            # dic, key=old_col, value=new_col. Some cols may be omitted and missing from the dic which are deemed irrelevant. Behaivour can be disabled by setting drop_debug_cols=False
            old2new_anno_cols = substitution_fn(c,cols_to_substitute,deducted_component_names[c])
            anno2final_cols[c.model] = list(old2new_anno_cols.values())
            new_cols = {**new_cols, **(old2new_anno_cols)}



        pipe.anno2final_cols = anno2final_cols
        cols_to_rename = list(new_cols.keys() )
        for k in cols_to_rename:
            # some cols might not exist because no annotations generated, so we need to double check it really exists
            if k not in df.columns: del new_cols[k]
        return df.rename(columns = new_cols)[set(new_cols.values()).union(set(stranger_cols))] if drop_debug_cols else df.rename(columns = new_cols)


    @staticmethod
    def get_final_output_cols_of_component(c,df,anno_2_ex):
        # get_final_output_cols_of_component(self.components[1], pretty_df, anno_2_ex_config)
        """Get's a list of all columns that have been derived in the pythonify procedure from the component c in dataframe df for anno_2_ex configs """
        og_output_col = c.info.spark_output_column_names[0]
        configs       = anno_2_ex[og_output_col]
        result_cols   = []
        if configs.get_annotator_type                 : result_cols.append(configs.output_col_prefix+'_types')
        if configs.get_result                         : result_cols.append(configs.output_col_prefix+'_results')
        if configs.get_begin or configs.get_positions : result_cols.append(configs.output_col_prefix+'_beginnings')
        if configs.get_end   or configs.get_positions : result_cols.append(configs.output_col_prefix+'_endings')
        if configs.get_embeds                         : result_cols.append(configs.output_col_prefix+'_embeddings')
        # find all metadata fields generated by compoent
        for col in df.columns :
            if 'meta_'+ configs.output_col_prefix in col:
                base_meta_prefix = 'meta_'+ configs.output_col_prefix
                meta_col_name = base_meta_prefix + col.split(base_meta_prefix)[-1]
                if meta_col_name in df.columns :
                    # special case for overlapping names with _
                    if col.split(base_meta_prefix)[-1].split('_')[1].isnumeric() and not c.info.outputs[0].split('_')[-1].isnumeric(): continue
                    if col.split(base_meta_prefix)[-1].split('_')[1].isnumeric() and c.info.outputs[0].split('_')[-1].isnumeric():
                        id1 = int(col.split(base_meta_prefix)[-1].split('_')[1])
                        id2 = int(c.info.outputs[0].split('_')[-1])
                        if id1 != id2 : continue
                    result_cols.append(meta_col_name)
                else : logger.info(f"Could not find meta col for c={c}, col={col}. Ommiting col..")
        return result_cols


    @staticmethod
    def deduct_component_names(pipe):
        """Deduct a meaningful name for Embeddings, classifiers, resolvesr, relation extractors, etc..
        Will return a dict that maps every Annotator Class to a String Name. If String_Name =='' that means, it can be omtited for naming and the
        unique_default name schema should be used,
        since that annotator is unique in the pipe
        """
        import nlu.pipe.col_substitution.name_deduction.name_deductable_annotators_OS as deductable_OS
        max_depth = 10
        result_names = {}
        for c in pipe.components :
            result_names[c]='UNIQUE' # assuemd uniqe, if not updated in followign steps
            is_always_name_deductable_component = False
            hc_deducted = False
            if pipe.has_licensed_components :
                import nlu.pipe.col_substitution.name_deduction.name_deductable_annotators_HC as deductable_HC
                if type(c.model) not in deductable_HC.name_deductable_HC and type(c.model) not in deductable_OS.name_deductable_OS: continue
                else : hc_deducted = True
                if type(c.model) in deductable_HC.always_name_deductable_HC: is_always_name_deductable_component=True

            if type(c.model) not in deductable_OS.name_deductable_OS and not hc_deducted : continue
            if type(c.model) in deductable_OS.always_name_deductable_OS: is_always_name_deductable_component=True

            same_components = []
            for other_c in pipe.components :
                if c is other_c: continue
                if c.info.type == other_c.info.type: same_components.append(other_c)
            if len(same_components) or is_always_name_deductable_component:
                # make sure each name is unique among the components of same type
                cur_depth = 1
                other_names = [ColSubstitutionUtils.deduct_name_from_nlu_ref_at_depth(other_c) for other_c in same_components]
                c_name = ColSubstitutionUtils.deduct_name_from_nlu_ref_at_depth(c)
                while c_name in other_names and cur_depth < max_depth:
                    cur_depth += 1
                    other_names = [ColSubstitutionUtils.deduct_name_from_nlu_ref_at_depth(other_c) for other_c in same_components]
                    c_name = ColSubstitutionUtils.deduct_name_from_nlu_ref_at_depth(c,cur_depth)
                result_names[c]=c_name
            else :
                result_names[c]='UNIQUE' # no name insertion required
        return result_names

    @staticmethod
    def deduct_name_from_nlu_ref_at_depth(c, depth=1):
        if isinstance(c.model, MarianTransformer): return c.info.nlu_ref.split('xx.')[-1].replace('marian.','')
        splits = c.info.nlu_ref.split('.')
        #remove all name irrelevant splits
        while  len(splits) >1 and (splits[0] in ColSubstitutionUtils.all_langs or splits[0] in ColSubstitutionUtils.cleanable_splits): splits.pop(0)
        if len(splits)==0:
            if isinstance(c.model,(NerDLModel,NerConverter)): return 'ner'
            return c.info.nlu_ref.replace("@","_")
        elif splits[0] == 'sentiment' and len(splits) == 1 : return 'UNIQUE'
        else : return '_'.join(splits[:depth])





from sparknlp.annotator import *


class SparkNLPBert:
    @staticmethod
    def get_default_model():
        return BertEmbeddings.pretrained() \
        .setInputCols("sentence", "token") \
        .setOutputCol("bert")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return BertEmbeddings.pretrained(name,language,bucket) \
            .setInputCols("sentence", "token") \
            .setOutputCol("bert")






from sparknlp.annotator import *
from nlu.pipe.viz.vis_utils_OS import VizUtilsOS
import random

class VizUtils():
    """Utils for interfacing with the Spark-NLP-Display lib"""
    @staticmethod
    def get_random(): return  random.randint(0,1333333333337)

    @staticmethod
    def infer_viz_type(pipe)->str:
        """For a given NLUPipeline, infers which visualizations are applicable. """
        if pipe.has_licensed_components :
            from nlu.pipe.viz.vis_utils_HC import VizUtilsHC
            return VizUtilsHC.infer_viz_licensed(pipe)
        else : return VizUtilsOS.infer_viz_open_source(pipe)


    @staticmethod
    def viz_OS(anno_res, pipe, viz_type,viz_colors,labels_to_viz,is_databricks_env,write_to_streamlit,streamlit_key):
        """Vizualize open source component"""
        streamlit_key = VizUtils.get_random() if streamlit_key == "RANDOM" else streamlit_key
        if   viz_type == 'ner' : return VizUtilsOS.viz_ner(anno_res, pipe,labels_to_viz,viz_colors,is_databricks_env,write_to_streamlit,streamlit_key)
        elif viz_type == 'dep' : return VizUtilsOS.viz_dep(anno_res, pipe,is_databricks_env,write_to_streamlit,streamlit_key)
        else : raise ValueError("Could not find applicable viz_type. Please make sure you specify either ner, dep, resolution, relation, assert or dep and have loaded corrosponding components")

    @staticmethod
    def viz_HC(anno_res, pipe, viz_type,viz_colors,labels_to_viz,is_databricks_env,write_to_streamlit):
        """Vizualize licensed component"""
        from nlu.pipe.viz.vis_utils_HC import VizUtilsHC
        if   viz_type == 'ner' : return  VizUtilsHC.viz_ner(anno_res, pipe,labels_to_viz,viz_colors,is_databricks_env,write_to_streamlit)
        elif viz_type == 'dep' : return  VizUtilsHC.viz_dep(anno_res, pipe,is_databricks_env,write_to_streamlit)
        elif viz_type == 'resolution' : return  VizUtilsHC.viz_resolution(anno_res, pipe,viz_colors,is_databricks_env,write_to_streamlit)
        elif viz_type == 'relation' : return  VizUtilsHC.viz_relation(anno_res, pipe,is_databricks_env,write_to_streamlit)
        elif viz_type == 'assert' : return  VizUtilsHC.viz_assertion(anno_res, pipe,viz_colors,is_databricks_env,write_to_streamlit)
        else : raise ValueError("Could not find applicable viz_type. Please make sure you specify either ner, dep, resolution, relation, assert or dep and have loaded corrosponding components")

"""Define whiche annotators model are definable by which vizualizer. There are 5 in total, 2 open source and 5 HC"""
# vizalbe_components_OC = {
#     'ner' : [NerConverter],
#     'dep' : [DependencyParserModel],
# }

# vizalbe_components_HC = {
#     'ner':[NerConverter,NerConverterInternal],
#     'resolution' : [SentenceEntityResolverModel, ChunkEntityResolverModel] ,
#     'relation'   : [RelationExtractionModel,RelationExtractionDLModel],
#     'assert'     : [AssertionDLModel,AssertionLogRegApproach],
#     'dep' : [DependencyParserModel],
# }





import unittest
from nlu import *

class TestSpam(unittest.TestCase):

    def test_spam_model(self):
        pipe = nlu.load('spam',verbose=True)
        df = pipe.predict(['I love pancaces. I hate Mondays', 'I love Fridays'], output_level='sentence')
        for c in df.columns: print(df[c])
        df = pipe.predict(['I love pancaces. I hate Mondays', 'I love Fridays'], output_level='document')
        for c in df.columns: print(df[c])



if __name__ == '__main__':
    unittest.main()




"""
This file contains methods to get pre-defined configurations for every annotator.
Extractor_resolver.py should be used to resolve SparkNLP Annotator classes to methods
in this file, which return the corrosponding configs that need to be passed to
the master_extractor() call.

This file is where all the in extractor_base_data_classes.py Dataclasses are combined with the
extractors defined in helper_extractor_methods.py.


"""

from nlu.pipe.extractors.extractor_base_data_classes import SparkNLPExtractor,SparkNLPExtractorConfig
from nlu.pipe.extractors.extractor_methods.helper_extractor_methods import *

"""
This file contains methods to get pre-defined configurations for every annotator.
Extractor_resolver.py should be used to resolve SparkNLP Annotator classes to methods 
in this file, which return the corrosponding configs that need to be passed to 
the master_extractor() call.

This file is where all the in extractor_base_data_classes.py Dataclasses are combined with the 
extractors defined in extractor_methods.py.

"""


def default_full_config(output_col_prefix='DEFAULT'):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_positions       = True,
        get_begin           = True,
        get_end             = True,
        get_embeds          = True,
        get_result          = True,
        get_meta            = True,
        get_full_meta       = True,
        get_annotator_type  = True,
        name                = 'default_full',
        description         = 'Default full configuration, keeps all data and gets all metadata fields',

    )
def default_document_config(output_col_prefix='document'):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        pop_result_list     = True,
        get_result          = True,
    )


def default_NER_config(output_col_prefix='NER'):
    """Extracts NER tokens withouth positions, just the IOB tags,confidences and classified tokens """
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        # get_result          = True,
        # get_meta            = True,
        meta_white_list     = ['confidence'],
        name                = 'default_ner',
        description         = 'NER with IOB tags and confidences for them',
    )

def meta_NER_config(output_col_prefix='NER'):
    """Extracts NER tokens withouth positions, just the IOB tags,confidences and classified tokens """
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        get_meta            = True,
        meta_white_list     = ['confidence'],
        name                = 'default_ner',
        description         = 'NER with IOB tags and confidences for them',
    )


def default_language_classifier_config(output_col_prefix='language'):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        get_meta            = True,
        get_full_meta       = True,
        pop_result_list     = True,
        name                = 'Only keep maximum language confidence',
        description         = 'Instead of returning the confidence for every language the Classifier was traiend on, only the maximum confidence will be returned',
        meta_data_extractor = SparkNLPExtractor(meta_extract_language_classifier_max_confidence,
                                                'Extract the maximum confidence from all classified languages and drop the others. TODO top k results',
                                                'Keep only top language confidence')
    )


def default_only_result_config(output_col_prefix):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        name                = 'Default result extractor',
        description         = 'Just gets the result field'
    )
def default_only_embedding_config(output_col_prefix):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_embeds          = True,
        name                = 'Default Embed extractor',
        description         = 'Just get the Embed field'
    )


def default_only_result_and_positions_config(output_col_prefix):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        get_positions       = True,

        name                = 'Positional result only default',
        description         = 'Get the result field and the positions'
    )


def default_sentiment_dl_config(output_col_prefix='sentiment_dl'):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        get_full_meta       = True,
        name                = 'Only keep maximum sentiment confidence ',
        description         = 'Instead of r eturning the confidence for Postive and Negative, only the confidence of the more likely class will be returned in the confidence column',
        meta_data_extractor = SparkNLPExtractor(extract_maximum_confidence,
                                                'Instead of returining positive/negative confidence, only the maximum confidence will be returned withouth sentence number reference.',
                                                'Maximum binary confidence')
    )

def default_lang_classifier_config(output_col_prefix='sentiment_dl'):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        get_full_meta       = True,
        pop_meta_list       = True,
        pop_result_list     = True,
        name                = 'default_lang_classifier_config',
        description         = 'Instead of returning the confidence for every language, just returns the confidence of the most likely language',
        meta_data_extractor = SparkNLPExtractor(extract_maximum_confidence,
                                                'Instead of returining positive/negative confidence, only the maximum confidence will be returned withouth sentence number reference.',
                                                'Maximum binary confidence')
    )


def default_sentiment_config(output_col_prefix='sentiment'):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        get_full_meta       = True,
        # pop_result_list     = True,
        name                = 'Only keep maximum sentiment confidence ',
        description         = 'Instead of returning the confidence for Postive and Negative, only the confidence of the more likely class will be returned in the confidence column',
        meta_data_extractor = SparkNLPExtractor(extract_maximum_confidence,
                                                'Instead of returining positive/negative confidence, only the maximum confidence will be returned withouth sentence number reference.',
                                                'Maximum binary confidence')
    )



def default_sentiment_vivk_config(output_col_prefix='vivk_sentiment'):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        get_full_meta       = True,
        # pop_result_list     = True,
        name                = 'Default sentiment vivk',
        description         = 'Get prediction confidence and the resulting label'
    )

def default_multi_classifier_dl_config(output_col_prefix='classifier_dl'):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        get_full_meta       = True,
        name                = 'default_classifier_dl',
        description         = 'Get all predicted confidences and labels',
        pop_never           = True,
        meta_data_extractor = SparkNLPExtractor(extract_maximum_confidence,
                                                'Instead returning confidence for each class, only return max confidence',
                                                'Max confidence')

    )

def default_classifier_dl_config(output_col_prefix='classifier_dl'):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        get_full_meta       = True,
        name                = 'default_classifier_dl',
        description         = 'Get all predicted confidences and labels',
        meta_data_extractor = SparkNLPExtractor(extract_maximum_confidence,
                                                'Instead returning confidence for each class, only return max confidence',
                                                'Max confidence')

    )


def default_tokenizer_config(output_col_prefix='token'):
    return default_only_result_config(output_col_prefix)

def default_POS_config(output_col_prefix='POS_tag'):
    return default_only_result_config(output_col_prefix)


def default_sentence_detector_DL_config(output_col_prefix='sentence'):
    return default_only_result_config(output_col_prefix)

def default_chunker_config(output_col_prefix='matched_chunk'):
    return default_only_result_config(output_col_prefix)




def default_T5_config(output_col_prefix='T5'):
    return default_only_result_config(output_col_prefix)
def default_marian_config(output_col_prefix='translated'):
    return default_only_result_config(output_col_prefix)


# EMBEDS
def default_sentence_embedding_config(output_col_prefix='sentence_embedding'):
    return default_only_embedding_config(output_col_prefix)

def default_chunk_embedding_config(output_col_prefix='chunk_embedding'):
    return default_only_embedding_config(output_col_prefix)

def default_word_embedding_config(output_col_prefix='word_embedding'):
    return default_only_embedding_config(output_col_prefix)

# TOKEN CLEANERS




def default_stopwords_config(output_col_prefix='stopwords_removed'):
    return default_only_result_config(output_col_prefix)
def default_lemma_config(output_col_prefix='lemma'):
    return default_only_result_config(output_col_prefix)
def default_stemm_config(output_col_prefix='stemm'):
    return default_only_result_config(output_col_prefix)
def default_norm_config(output_col_prefix='norm'):
    return default_only_result_config(output_col_prefix)
def default_norm_document_config(output_col_prefix='norm_document'):
    return default_only_result_config(output_col_prefix)

def default_sentence_detector_config(output_col_prefix='sentence'):
    return default_only_result_config(output_col_prefix)




def default_dep_untyped_config(output_col_prefix='dependency_untyped'):
    return default_only_result_config(output_col_prefix)

def default_dep_typed_config(output_col_prefix='dependency_typed'):
    return default_only_result_config(output_col_prefix)




def default_spell_norvig_config(output_col_prefix='spell_checked'):
    return default_only_result_config(output_col_prefix)

def default_spell_context_config(output_col_prefix='spell_checked'):
    return default_only_result_config(output_col_prefix)

def default_spell_symmetric_config(output_col_prefix='spell_checked'):
    return default_only_result_config(output_col_prefix)
def default_ngram_config(output_col_prefix='ngram'):
    return default_only_result_config(output_col_prefix)

def default_word_segmenter_config(output_col_prefix='words_segmented'):
    return default_only_result_config(output_col_prefix)


def default_chunk_config(output_col_prefix='matched_chunk'):
    return default_only_result_config(output_col_prefix)


def default_yake_config(output_col_prefix='keywords'):
    """Extracts YAKE keywords with confidences """
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        name                = 'default_yake',
        get_meta            = True,
        meta_white_list     = ['score'],
        description         = 'Get all keywords and their confidences',
        pop_never           = True
    )





def default_ner_converter_config(output_col_prefix='ner_chunk'):
    """Extracts the Entity Labels, which are derived from the IOB Tags """
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        name                = 'default_ner',
        get_meta            = True,
        meta_white_list     = ['entity','confidence'],
        description         = 'Converts IOB-NER representation into entity representation and generates confidences for the entire entity chunk',
    )

def default_doc2chunk_config(output_col_prefix='doc2chunk'):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = False,
        name                = 'doc2chunk',
        get_meta            = False,
        description         = 'Converts Doc type col to chunk aka entity type',
    )





import unittest
from nlu import *
class TestCyber(unittest.TestCase):

    def test_snips_classifer_model(self):
        pipe = nlu.load('en.classify.snips',verbose=True)
        df = pipe.predict(['I love pancaces. I hate Mondays', 'I love Fridays'])
        print(df.columns)
        for c in df.columns:print(c,df[c])
    def test_snips_ner_model(self):
        pipe = nlu.load('en.ner.snips',verbose=True)
        df = pipe.predict(['I love pancaces. I hate Mondays', 'I love Fridays'])
        print(df.columns)
        for c in df.columns:print(c,df[c])



if __name__ == '__main__':
    unittest.main()




import unittest
from tests.test_utils import get_sample_pdf_with_labels, get_sample_pdf, get_sample_sdf, get_sample_pdf_with_extra_cols, get_sample_pdf_with_no_text_col ,get_sample_spark_dataframe
from nlu import *

class MatchTests(unittest.TestCase):

    def test_pattern_matcher(self):
        pass
        pipe = nlu.load('match.pattern', verbose=True )
        df = pipe.predict('2020 was a crazy year but wait for October 1. 2020')
        for c in df.columns: print(df[c])


    def test_chunk_matcher(self):
        pass
        pipe = nlu.load('match.chunks', verbose=True )
        df = pipe.predict('2020 was a crazy year but wait for October 1. 2020')
        for c in df.columns: print(df[c])

    def download_entities_files(self):
        import urllib2
        response = urllib2.urlopen('https://wordpress.org/plugins/about/readme.txt')
        data = response.read()
        filename = "readme.txt"
        file_ = open(filename, 'w')
        file_.write(data)
        file_.close()


    def test_text_matcher(self):
        p = '/home/ckl/Documents/freelance/jsl/nlu/nlu4realgit2/tmp/trasgh/entities.txt'
        pipe = nlu.load('match.text', verbose=True )
        pipe['text_matcher'].setEntities(p)
        df = pipe.predict('2020 was a crazy year but wait for October 1. 2020')
        for c in df.columns: print(df[c])
    def test_regex_matcher(self):
        p='/home/ckl/Documents/freelance/jsl/nlu/nlu4realgit2/tmp/trasgh/rulesd.txt'
        pipe = nlu.load('match.regex', verbose=True )
        pipe['regex_matcher'].setExternalRules(path=p, delimiter=',')
        df = pipe.predict('2020 was a crazy year but wait for October 1. 2020')
        for c in df.columns: print(df[c])
    def test_date_matcher(self):
        pipe = nlu.load('match.date', verbose=True )
        df = pipe.predict('2020 was a crazy year but wait for October 1. 2020')
        for c in df.columns: print(df[c])

if __name__ == '__main__':
    unittest.main()




import unittest
from tests.test_utils import get_sample_pdf_with_labels, get_sample_pdf, get_sample_sdf, get_sample_pdf_with_extra_cols, get_sample_pdf_with_no_text_col ,get_sample_spark_dataframe
from nlu import *

class TestLem(unittest.TestCase):

    def test_stem_pipe(self):
        pipe = nlu.load('lemma', verbose=True )
        df = pipe.predict('HELLO WORLD! How are YOU!?!@', output_level='sentence',drop_irrelevant_cols=False, metadata=True, )
        for c in df.columns: print(df[c])


if __name__ == '__main__':
    unittest.main()





from nlu.pipe.extractors.extractor_base_data_classes import SparkNLPExtractor,SparkNLPExtractorConfig
from nlu.pipe.extractors.extractor_methods.helper_extractor_methods import *
from nlu.pipe.extractors.extractor_methods.base_extractor_methods import *

"""
This file contains methods to get pre-defined configurations for every annotator.
Extractor_resolver.py should be used to resolve SparkNLP Annotator classes to methods 
in this file, which return the corrosponding configs that need to be passed to 
the master_extractor() call.

This file is where all the in extractor_base_data_classes.py Dataclasses are combined with the 
extractors defined in extractor_methods.py.

"""

def default_get_nothing(output_col_prefix):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        name                = 'nothing_extractor',
        description         = 'Extracts nothing. Useful for annotators with irrelevant data'
    )
def default_only_result_config(output_col_prefix):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        name                = 'Default result extractor',
        description         = 'Just gets the result field'
    )

def default_full_config(output_col_prefix='DEFAULT'):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_positions       = True,
        get_begin           = True,
        get_end             = True,
        get_embeds          = True,
        get_result          = True,
        get_meta            = True,
        get_full_meta       = True,
        get_annotator_type  = True,
        name                = 'default_full',
        description         = 'Default full configuration, keeps all data and gets all metadata fields',

    )




def default_NER_converter_licensed_config(output_col_prefix='entities'):
    """Extracts NER tokens withouth positions, just the converted IOB tags,confidences and classified tokens """
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        get_meta            = True,
        meta_white_list     = ['entity','confidence'], #sentence, chunk
        name                = 'default_ner',
        description         = 'Converts IOB-NER representation into entity representation and generates confidences for the entire entity chunk',
    )

def default_chunk_resolution_config(output_col_prefix='resolved_entities'):
    """Extracts NER tokens withouth positions, just the IOB tags,confidences and classified tokens """
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        get_meta            = True,
        meta_white_list     = ['confidence','resolved_text'], #sentence, chunk
        name                = 'default_ner',
        description         = 'Converts IOB-NER representation into entity representation and generates confidences for the entire entity chunk',
    )

def default_relation_extraction_positional_config(output_col_prefix='extracted_relations'):
    """Extracts NER tokens withouth positions, just the IOB tags,confidences and classified tokens """
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        meta_white_list     = [],
        get_meta            = True,
        meta_black_list     = ['entity1_begin','entity2_begin','entity1_end','entity2_end','entity1','entity2','chunk1','chunk2',],
        name                = 'default_relation_extraction',
        description         = 'Get relation extraction result and all metadata, positions of entities excluded',
    )


def default_relation_extraction_config(output_col_prefix='extracted_relations'):
    """Extracts NER tokens withouth positions, just the IOB tags,confidences and classified tokens """
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        meta_white_list     = [],
        get_meta            = True,
        meta_black_list     = ['entity1','entity2','chunk1','chunk2',],
        name                = 'default_relation_extraction',
        description         = 'Get relation extraction result and all metadata, positions of entities excluded',
    )




def default_de_identification_config(output_col_prefix='de_identified'):
    """Extracts NER tokens withouth positions, just the IOB tags,confidences and classified tokens """
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        name                = 'positional_relation_extraction',
        description         = 'Get relation extraction result and all metadata, which will include positions of entities chunks',
    )


def default_assertion_config(output_col_prefix='assertion'):
    """Extracts NER tokens withouth positions, just the IOB tags,confidences and classified tokens """
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        name                = 'default_assertion_extraction',
        get_meta            = True,
        meta_white_list     = ['confidence'],
        description         = 'Gets the assertion result and confidence',
    )

def default_ner_config(output_col_prefix='med_ner'):
    return default_only_result_config(output_col_prefix)

def default_ner_config(output_col_prefix='med_ner'):
    return default_get_nothing(output_col_prefix)


def default_feature_assembler_config(output_col_prefix='feature_assembler'):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = False,
        name                = 'features_assembled',
        get_meta            = False,
        description         = 'Gets nothing',
    )

def default_generic_classifier_config(output_col_prefix='generic_classifier'):
    return SparkNLPExtractorConfig(
        output_col_prefix   = output_col_prefix,
        get_result          = True,
        name                = 'generic_classifier',
        get_meta            = True,
        meta_white_list     = ['confidence'],
        description         = 'Gets the  result and confidence',
    )


from sparknlp.annotator import *

class SparkNLPElmo:
    @staticmethod
    def get_default_model():
        return ElmoEmbeddings.pretrained() \
        .setInputCols("sentence", "token") \
        .setOutputCol("word_embeddings")



    @staticmethod
    def get_pretrained_model(name, language):
        return ElmoEmbeddings.pretrained(name,language) \
            .setInputCols("sentence", "token") \
            .setOutputCol("word_embeddings")




import logging
logger = logging.getLogger('nlu')

from nlu.pipe.extractors.extraction_resolver_OS import OS_anno2config
from nlu.pipe.extractors.extractor_methods.base_extractor_methods import *
from nlu.pipe.pipe_logic import PipeUtils

import nlu
import nlu.pipe.pipe_components
import sparknlp
from  typing import List,Union

from sparknlp.base import *
from sparknlp.base import LightPipeline
from sparknlp.annotator import *
import pyspark
import pandas as pd
import numpy as np
from pyspark.sql.types import StructType,StructField, StringType
from nlu.pipe.component_resolution import extract_classifier_metadata_from_nlu_ref
from nlu.pipe.utils.storage_ref_utils import StorageRefUtils
from nlu.pipe.utils.component_utils import ComponentUtils
from nlu.pipe.utils.output_level_resolution_utils import OutputLevelUtils
from nlu.pipe.utils.data_conversion_utils import DataConversionUtils
from nlu.utils.environment.env_utils import is_running_in_databricks
from nlu.pipe.col_substitution.col_name_substitution_utils import ColSubstitutionUtils
class BasePipe(dict):
    # we inherhit from dict so the pipe is indexable and we have a nice shortcut for accessing the spark nlp model
    def __init__(self):
        self.nlu_ref=''
        self.raw_text_column = 'text'
        self.raw_text_matrix_slice = 1  # place holder for getting text from matrix
        self.spark_nlp_pipe = None
        self.has_trainable_components = False
        self.needs_fitting = True
        self.is_fitted = False
        self.output_positions = False  # Wether to putput positions of Features in the final output. E.x. positions of tokens, entities, dependencies etc.. inside of the input document.
        self.output_level = ''  # either document, chunk, sentence, token
        self.output_different_levels = True
        self.light_pipe_configured = False
        self.spark_non_light_transformer_pipe = None
        self.components = []  # orderd list of nlu_component objects
        self.output_datatype = 'pandas'  # What data type should be returned after predict either spark, pandas, modin, numpy, string or array
        self.lang = 'en'
        self.spark_transformer_pipe = None
        self.spark_estimator_pipe_pipe = None
        self.has_licensed_components = False
    def isInstanceOfNlpClassifer(self, model):
        '''
        Check for a given Spark NLP model if it is an instance of a classifier , either approach or already fitted transformer will return true
        This is used to configured the input/output columns based on the inputs
        :param model: the model to check
        :return: True if it is one of the following classes : (ClassifierDLModel,ClassifierDLModel,MultiClassifierDLModel,MultiClassifierDLApproach,SentimentDLModel,SentimentDLApproach) )
        '''
        return isinstance(model, (
            ClassifierDLModel, ClassifierDLModel, MultiClassifierDLModel, MultiClassifierDLApproach, SentimentDLModel,
            SentimentDLApproach))
    def configure_outputs(self, component, nlu_reference, name_to_add=''):
        '''
        Configure output column names of classifiers from category to something more meaningful
        Name should be Name of classifier, based on NLU reference.
        Duplicate names will be resolved by appending suffix "_i" to column name, based on how often we encounterd duplicate errors
        This updates component infos accordingly
        :param component: classifier component for which the output columns to  configured
        :param nlu_reference: nlu reference from which is component stemmed
        :return: None
        '''
        if nlu_reference == 'default_name' : return
        nlu_reference = nlu_reference.replace('train.', '')
        model_meta = extract_classifier_metadata_from_nlu_ref(nlu_reference)
        can_use_name = False
        new_output_name = model_meta[0]
        i = 0
        while can_use_name == False:
            can_use_name = True
            for c in self.components:
                if new_output_name in c.info.spark_input_column_names + c.info.spark_output_column_names and c.info.name != component.info.name:
                    can_use_name = False
        if can_use_name == False:
            new_output_name = new_output_name + '_' + str(i)
            i += 1
        # classifiers always have just 1 output col
        logger.info(f"Configured output columns name to {new_output_name} for classifier in {nlu_reference}")
        component.model.setOutputCol(new_output_name)
        component.info.spark_output_column_names = [new_output_name]
    def add(self, component, nlu_reference="default_name", pretrained_pipe_component=False, name_to_add=''):
        '''

        :param component:
        :param nlu_reference: NLU references, passed for components that are used specified and not automatically generate by NLU
        :return:
        '''
        if hasattr(component.info,'nlu_ref'): nlu_reference = component.info.nlu_ref
        # self.nlu_reference = component.info.nlu_ref if component.info
        self.components.append(component)
        # ensure that input/output cols are properly set
        # Spark NLP model reference shortcut

        name = component.info.name.replace(' ', '').replace('train.', '')

        if StorageRefUtils.has_storage_ref(component):
            name = name +'@' + StorageRefUtils.extract_storage_ref(component)
        logger.info(f"Adding {name} to internal pipe")

        # Configure output column names of classifiers from category to something more meaningful
        if self.isInstanceOfNlpClassifer(component.model): self.configure_outputs(component, nlu_reference)

        if name_to_add == '':
            # Add Component as self.index and in attributes
            if 'embed' in component.info.type and nlu_reference not in self.keys() and not pretrained_pipe_component:
                self[name] = component.model
            elif name not in self.keys():
                if not hasattr(component.info,'nlu_ref'):  component.info.nlu_ref  = nlu_reference
                self[name] = component.model
            else:
                nlu_identifier = ComponentUtils.get_nlu_ref_identifier(component)
                component.info.nlu_ref = nlu_reference
                self[component.info.name +"@"+ nlu_identifier] = component.model
        else :self[name_to_add] = component.model
class NLUPipeline(BasePipe):
    def __init__(self):
        super().__init__()
        """ Initializes a pretrained pipeline         """
        self.spark = sparknlp.start()
        self.provider = 'sparknlp'
        self.pipe_ready = False  # ready when we have created a spark df
        # The NLU pipeline uses  types of Spark NLP annotators to identify how to handle different columns
    def get_sample_spark_dataframe(self):
        data = {"text": ['This day sucks', 'I love this day', 'I dont like Sami']}
        text_df = pd.DataFrame(data)
        return sparknlp.start().createDataFrame(data=text_df)
    def verify_all_labels_exist(self,dataset):
        return 'y'  in dataset.columns or 'label' in dataset.columns or 'labels' in dataset.columns
    def fit(self, dataset=None, dataset_path=None, label_seperator=','):
        '''
        if dataset is  string with '/' in it, its dataset path!
        Converts the input Pandas Dataframe into a Spark Dataframe and trains a model on it.
        :param dataset: The pandas dataset to train on, should have a y column for label and 'text' column for text features
        :param dataset_path: Path to a CONLL2013 format dataset. It will be read for NER and POS training.
        :param label_seperator: If multi_classifier is trained, this seperator is used to split the elements into an Array column for Pyspark
        :return: A nlu pipeline with models fitted.
        '''
        self.is_fitted = True
        stages = []
        for component in self.components:stages.append(component.model)
        self.spark_estimator_pipe = Pipeline(stages=stages)

        if dataset_path != None and 'ner' in self.nlu_ref:
            from sparknlp.training import CoNLL
            s_df = CoNLL().readDataset(self.spark,path=dataset_path, )
            self.spark_transformer_pipe = self.spark_estimator_pipe.fit(s_df.withColumnRenamed('label','y'))

        elif dataset_path != None and 'pos' in self.nlu_ref:
            from sparknlp.training import POS
            s_df = POS().readDataset(self.spark,path=dataset_path,delimiter=label_seperator,outputPosCol="y",outputDocumentCol="document",outputTextCol="text")
            self.spark_transformer_pipe = self.spark_estimator_pipe.fit(s_df)
        elif isinstance(dataset,pd.DataFrame) and 'multi' in  self.nlu_ref:
            schema = StructType([
                StructField("y", StringType(), True),
                StructField("text", StringType(), True)
            ])
            from pyspark.sql import functions as F
            df = self.spark.createDataFrame(data=dataset).withColumn('y',F.split('y',label_seperator))
            # df = self.spark.createDataFrame(data=dataset, schema=schema).withColumn('y',F.split('y',label_seperator))
            # df = self.spark.createDataFrame(dataset)
            self.spark_transformer_pipe = self.spark_estimator_pipe.fit(df)

        elif isinstance(dataset,pd.DataFrame):
            if not self.verify_all_labels_exist(dataset) : raise ValueError(f"Could not detect label in provided columns={dataset.columns}\nMake sure a column named label, labels or y exists in your dataset.")
            if self.has_licensed_components :
                for c in self.components :
                    if c.info.name == 'feature_assembler':
                        vector_assembler_input_cols=[c for c in dataset.columns if c !='text' and c !='y' and c!='label' and c!='labels']
                        c.model.setInputCols(vector_assembler_input_cols)
                        # c.model.info.spark_input_column_names = vector_assembler_input_cols


            # self.spark_transformer_pipe = self.spark_estimator_pipe.fit(DataConversionUtils.pdf_to_sdf(dataset,self.spark)[0])

            # TODO GOD DAMMIT AT NOTATION MAKES WIERD SQL ERRORS APPEAR HERE!!! WE GOTTA REMOVE IT and subitutite it with smth different......
            for c in self.components:
                for inp in c.info.spark_input_column_names:
                    if 'chunk_embedding' in inp :
                        c.info.spark_input_column_names.remove(inp)
                        c.info.spark_input_column_names.append(inp.replace('@',"___"))
                        c.model.setInputCols(c.info.spark_input_column_names)
                for out in c.info.spark_output_column_names:
                    if 'chunk_embedding' in out :
                        c.info.spark_output_column_names.remove(out)
                        c.info.spark_output_column_names.append(out.replace('@',"___"))
                        c.model.setOutputCol(c.info.spark_output_column_names[0])

            stages = []
            for component in self.components:stages.append(component.model)
            ## TODO SET STORAGE REF ON FITTED ANNOTATORS!!!!especially resoluton...
            self.spark_estimator_pipe = Pipeline(stages=stages)
            self.spark_transformer_pipe = self.spark_estimator_pipe.fit(DataConversionUtils.pdf_to_sdf(dataset,self.spark)[0])

        elif isinstance(dataset,pyspark.sql.DataFrame) :
            if not self.verify_all_labels_exist(dataset) : raise ValueError(f"Could not detect label in provided columns={dataset.columns}\nMake sure a column named label, labels or y exists in your dataset.")
            self.spark_transformer_pipe = self.spark_estimator_pipe.fit(DataConversionUtils.sdf_to_sdf(dataset,self.spark))

        else :
            # fit on empty dataframe since no data provided
            logger.info('Fitting on empty Dataframe, could not infer correct training method. This is intended for non-trainable pipelines.')
            self.spark_transformer_pipe = self.spark_estimator_pipe.fit(self.get_sample_spark_dataframe())
        return self

    def get_annotator_extraction_configs(self,full_meta,c_level_mapping,positions):

        """Search first OC namespace and if not found the HC Namespace for each Annotator Class in pipeline and get corrosponding config
        Returns a dictionary of methods, where keys are column names values are methods  that are applied to extract and represent the data in these
        these columns in a more pythonic and panda-esque way
        """
        # tood doc level annos and same level annos can be popped always.
        anno_2_ex_config = {}
        for c in self.components:
            if type(c.model) in OS_anno2config.keys():
                if OS_anno2config[type(c.model)]['default'] == '' or full_meta:
                    if not full_meta : logger.info(f'could not find default configs, using full default for model ={c.model}')
                    anno_2_ex_config[c.info.spark_output_column_names[0]] = OS_anno2config[type(c.model)]['default_full'](output_col_prefix=c.info.outputs[0])
                else :
                    anno_2_ex_config[c.info.spark_output_column_names[0]] = OS_anno2config[type(c.model)]['default'](output_col_prefix=c.info.outputs[0])
            else:
                from nlu.pipe.extractors.extraction_resolver_HC import HC_anno2config
                if HC_anno2config[type(c.model)]['default'] == '' or full_meta :
                    if not full_meta : logger.info(f'could not find default configs in hc resolver space, using full default for model ={c.model}')
                    anno_2_ex_config[c.info.spark_output_column_names[0]] = HC_anno2config[type(c.model)]['default_full'](output_col_prefix=c.info.outputs[0])
                else :
                    anno_2_ex_config[c.info.spark_output_column_names[0]] = HC_anno2config[type(c.model)]['default'](output_col_prefix=c.info.outputs[0])

            if c_level_mapping[c] == 'document' and  not anno_2_ex_config[c.info.spark_output_column_names[0]].pop_never :
                anno_2_ex_config[c.info.spark_output_column_names[0]].pop_meta_list = True
                anno_2_ex_config[c.info.spark_output_column_names[0]].pop_result_list = True
            # if c_level_mapping[c] == 'relation' and  not anno_2_ex_config[c.info.spark_output_column_names[0]].pop_never :
            if positions : anno_2_ex_config[c.info.spark_output_column_names[0]].get_positions = True
            else :
                anno_2_ex_config[c.info.spark_output_column_names[0]].get_begin           = False
                anno_2_ex_config[c.info.spark_output_column_names[0]].get_end             = False
                anno_2_ex_config[c.info.spark_output_column_names[0]].get_positions = False

        return anno_2_ex_config
    def unpack_and_apply_extractors(self,sdf:pyspark.sql.DataFrame, keep_stranger_features=True, stranger_features=[],anno_2_ex_config={})-> pd.DataFrame:
        """1. Unpack SDF to PDF with Spark NLP Annotator Dictionaries
           2. Get the extractor configs for the corrosponding Annotator classes
           3. Apply The extractor configs with the extractor methods to each column and merge back with zip/explode"""
        unpack_df = sdf.toPandas().applymap(extract_pyspark_rows)
        return apply_extractors_and_merge(unpack_df,anno_2_ex_config, keep_stranger_features,stranger_features)
    def pythonify_spark_dataframe(self, processed,
                                  keep_stranger_features=True,
                                  stranger_features=[],
                                  drop_irrelevant_cols=True,
                                  output_metadata=False,
                                  positions = False
                                  ):
        '''
        This functions takes in a spark dataframe with Spark NLP annotations in it and transforms it into a Pandas Dataframe with common feature types for further NLP/NLU downstream tasks.
        It will recylce Indexes from Pandas DataFrames and Series if they exist, otherwise a custom id column will be created which is used as inex later on
            It does this by performing the following consecutive steps :
                1. Select columns to explode
                2. Select columns to keep
                3. Rename columns
                4. Create Pandas Dataframe object


        :param processed: Spark dataframe which an NLU pipeline has transformed
        :param output_level: The output level at which returned pandas Dataframe should be
        :param get_different_level_output:  Wheter to get features from different levels
        :param keep_stranger_features : Wether to keep additional features from the input DF when generating the output DF or if they should be discarded for the final output DF
        :param stranger_features: A list of features which are not known to NLU and inside of the input DF.
                                    Basically all columns, which are not named 'text' in the input.
                                    If keep_stranger_features== True, then these features will be exploded, if output_level == DOCUMENt, otherwise they will not be exploded
        :param output_metadata: Wether to keep or drop additional metadataf or predictions, like prediction confidence
        :return: Pandas dataframe which easy accessable features
        '''
        stranger_features += ['origin_index']

        if self.output_level == '' : self.output_level = OutputLevelUtils.infer_output_level(self)
        c_level_mapping =OutputLevelUtils.get_output_level_mapping_by_component(self)

        anno_2_ex_config                                                          = self.get_annotator_extraction_configs(output_metadata,c_level_mapping,positions)
        # TODO UNPACK MAKES MULTI_LAEL CLASS TRAINED GO OOOOOOOOOOOOOOOOOOOOOOOOOOOM
        pretty_df = self.unpack_and_apply_extractors(processed, keep_stranger_features, stranger_features,anno_2_ex_config)
        col2output_level,same_output_level,not_same_output_level                  = OutputLevelUtils.get_output_level_mappings(self,pretty_df,anno_2_ex_config)
        logger.info(f"Extracting for same_level_cols = {same_output_level}\nand different_output_level_cols = {not_same_output_level}")
        pretty_df = zip_and_explode(pretty_df, same_output_level, not_same_output_level, self.output_level)
        pretty_df = self.convert_embeddings_to_np(pretty_df)
        pretty_df = ColSubstitutionUtils.substitute_col_names(pretty_df,anno_2_ex_config,self, stranger_features)
        pretty_df = pretty_df.loc[:,~pretty_df.columns.duplicated()]
        if  drop_irrelevant_cols :  pretty_df = pretty_df[self.drop_irrelevant_cols(list(pretty_df.columns))]
        return pretty_df



    def convert_embeddings_to_np(self, pdf):
        '''
        convert all the columns in a pandas df to numpy
        :param pdf: Pandas Dataframe whose embedding column will be converted to numpy array objects
        :return:
        '''

        for col in pdf.columns:
            if 'embed' in col:
                pdf[col] = pdf[col].apply(lambda x: np.array(x))
        return pdf
    def finalize_return_datatype(self, df):
        '''
        Take in a Spark dataframe with only relevant columns remaining.
        Depending on what value is set in self.output_datatype, this method will cast the final SDF into Pandas/Spark/Numpy/Modin/List objects
        :param df:
        :return: The predicted Data as datatype dependign on self.output_datatype
        '''
        if self.output_datatype == 'spark':
            return df
        elif self.output_datatype == 'pandas':
            return df
        elif self.output_datatype == 'modin':
            import modin.pandas as mpd
            return mpd.DataFrame(df)
        elif self.output_datatype == 'pandas_series':
            return df
        elif self.output_datatype == 'modin_series':
            import modin.pandas as mpd
            return mpd.DataFrame(df)
        elif self.output_datatype == 'numpy':
            return df.to_numpy()
        return df
    def drop_irrelevant_cols(self, cols,  keep_origin_index=False):
        '''
        Takes in a list of column names removes the elements which are irrelevant to the current output level.
        This will be run before returning the final df
        Drop column candidates are document, sentence, token, chunk.
        columns which are NOT AT THE SAME output level will be dropped
        :param cols:  list of column names in the df
        :return: list of columns with the irrelevant names removed
        '''
        if self.output_level == 'token':
            if 'document' in cols: cols.remove('document')
            if 'chunk' in cols: cols.remove('chunk')
            if 'sentence' in cols: cols.remove('sentence')
        if self.output_level == 'sentence':
            if 'token' in cols: cols.remove('token')
            if 'chunk' in cols: cols.remove('chunk')
            if 'document' in cols: cols.remove('document')
        if self.output_level == 'chunk':
            # if 'document' in cols: cols.remove('document')
            if 'token' in cols: cols.remove('token')
            if 'sentence' in cols: cols.remove('sentence')
        if self.output_level == 'document':
            if 'token' in cols: cols.remove('token')
            if 'chunk' in cols: cols.remove('chunk')
            if 'sentence' in cols: cols.remove('sentence')
        if self.output_level == 'relation':
            if 'token' in cols: cols.remove('token')
            if 'chunk' in cols: cols.remove('chunk')
            if 'sentence' in cols: cols.remove('sentence')
        if keep_origin_index == False and 'origin_index' in cols : cols.remove('origin_index')
        return cols
    def configure_light_pipe_usage(self, data_instances, use_multi=True, force = False):
        logger.info("Configuring Light Pipeline Usage")
        if data_instances > 50000 or use_multi == False:
            logger.info("Disabling light pipeline")
            if  not self.is_fitted :self.fit()
        else:
            if self.light_pipe_configured == False or force and not isinstance(self.spark_transformer_pipe, LightPipeline):
                self.light_pipe_configured = True
                logger.info("Enabling light pipeline")
                self.spark_transformer_pipe = LightPipeline(self.spark_transformer_pipe)

    def check_if_sentence_level_requirements_met(self):
        '''
        Check if the pipeline currently has an annotator that generate sentence col as output. If not, return False
        :return:
        '''

        for c in self.components:
            if 'sentence' in c.info.spark_output_column_names : return True
        return False
    def add_missing_sentence_component(self):
        '''
        Add Sentence Detector to pipeline and Run it thorugh the Query Verifiyer again.
        :return: None
        '''

    def write_nlu_pipe_info(self,path):
        '''
        Writes all information required to load a NLU pipeline from disk to path
        :param path: path where to store the nlu_info.json
        :return: True if success, False if failure
        '''
        import os
        f = open(os.path.join(path,'nlu_info.txt'), "w")
        f.write(self.nlu_ref)
        f.close()
        return True
    def add_missing_component_if_missing_for_output_level(self):
        '''
        Check that for currently configured self.output_level one annotator for that level exists, i.e a Sentence Detetor for outpul tevel sentence, Tokenizer for level token etc..
        :return: None
        '''
        if self.output_level =='sentence':
            if self.check_if_sentence_level_requirements_met(): return
            else :
                logger.info('Adding missing sentence Dependency because it is missing for outputlevel=Sentence')
                self.add_missing_sentence_component()
    def save(self, path, component='entire_pipeline', overwrite=False):
        if nlu.is_running_in_databricks() :
            if path.startswith('/dbfs/') or path.startswith('dbfs/'):
                nlu_path = path
                if path.startswith('/dbfs/'):
                    nlp_path =  path.replace('/dbfs','')
                else :
                    nlp_path =  path.replace('dbfs','')
            else :
                nlu_path = 'dbfs/' + path
                if path.startswith('/') : nlp_path = path
                else : nlp_path = '/' + path

            if not self.is_fitted and self.has_trainable_components:
                self.fit()
                self.is_fitted = True
            if component == 'entire_pipeline':
                self.spark_transformer_pipe.save(nlp_path)
                self.write_nlu_pipe_info(nlu_path)
        if overwrite and not nlu.is_running_in_databricks():
            import shutil
            shutil.rmtree(path,ignore_errors=True)
        if not self.is_fitted :
            self.fit()
            self.is_fitted = True
        if component == 'entire_pipeline':
            if isinstance(self.spark_transformer_pipe, LightPipeline):
                self.spark_transformer_pipe.pipeline_model.save(path)
            else: self.spark_transformer_pipe.save(path)
        else:
            if component in self.keys():
                self[component].save(path)
        print(f'Stored model in {path}')

    def predict(self,
                data,
                output_level='',
                positions=False,
                keep_stranger_features=True,
                metadata=False,
                multithread=True,
                drop_irrelevant_cols=True,
                return_spark_df = False):
        '''
        Annotates a Pandas Dataframe/Pandas Series/Numpy Array/Spark DataFrame/Python List strings /Python String

        :param data: Data to predict on
        :param output_level: output level, either document/sentence/chunk/token
        :param positions: wether to output indexes that map predictions back to position in origin string
        :param keep_stranger_features: wether to keep columns in the dataframe that are not generated by pandas. I.e. when you s a dataframe with 10 columns and only one of them is named text, the returned dataframe will only contain the text column when set to false
        :param metadata: wether to keep additonal metadata in final df or not like confidiences of every possible class for preidctions.
        :param multithread: Whether to use multithreading based lightpipeline. In some cases, this may cause errors.
        :param drop_irellevant_cols: Wether to drop cols of different output levels, i.e. when predicting token level and dro_irrelevant_cols = True then chunk, sentence and Doc will be dropped
        :param return_spark_df: Prediction results will be returned right after transforming with the Spark NLP pipeline
        :return:
        '''
        if output_level != '': self.output_level = output_level
        if output_level == 'sentence' or output_level == 'document':self.components = PipeUtils.configure_component_output_levels(self)
        if output_level in ['token', 'chunk','relation']:self.components = PipeUtils.configure_component_output_levels(self,'document')

        try:
            #1. Convert data to Spark DF
            data, stranger_features, output_datatype = DataConversionUtils.to_spark_df(data, self.spark, self.raw_text_column)

            #2. configure Lightpipline usage
            self.configure_light_pipe_usage(data.count(), multithread)
            # 3. Fit
            if not self.is_fitted :
                if self.has_trainable_components:self.fit(data)
                else : self.fit()

            #4. Apply Spark Pipeline
            data = self.spark_transformer_pipe.transform(data)
            #5. Convert resulting spark DF into nicer format and by default into pandas.
            if return_spark_df : return data  # Returns RAW  Spark Dataframe result of pipe prediction
            return self.pythonify_spark_dataframe(data,
                                                  keep_stranger_features=keep_stranger_features,
                                                  stranger_features=stranger_features,
                                                   output_metadata=metadata,
                                                  drop_irrelevant_cols=drop_irrelevant_cols,
                                                  positions=positions
                                                  )





        except Exception as err :

            import sys
            if multithread == True:
                logger.warning("Multithreaded mode failed. trying to predict again with non multithreaded mode ")
                # return self.predict(data, output_level=output_level, positions=positions,keep_stranger_features=keep_stranger_features, metadata=metadata, multithread=False,return_spark_df=return_spark_df)
                return  self.predict(data=data,output_level=output_level,positions=positions,keep_stranger_features=keep_stranger_features,metadata=metadata,multithread=False,drop_irrelevant_cols=drop_irrelevant_cols,return_spark_df=return_spark_df)
            else: self.print_exception_err(err)


    def print_info(self,minimal=True):
        '''
        Print out information about every component currently loaded in the pipe and their configurable parameters.
        If minimal is false, all Spark NLP Model parameters will be printed, including output/label/input cols and other attributes a NLU user should not touch. Useful for debugging.
        :return: None
        '''
        print('The following parameters are configurable for this NLU pipeline (You can copy paste the examples) :')
        # list of tuples, where first element is component name and second element is list of param tuples, all ready formatted for printing
        all_outputs = []
        ## TODO CONDINOTAL LOOP either on approaches or transformers
        iterable = None
        for i, component_key in enumerate(self.keys()):
            s = ">>> pipe['" + component_key + "'] has settable params:"
            p_map = self[component_key].extractParamMap()

            component_outputs = []
            max_len = 0
            for key in p_map.keys():
                if minimal:
                    if "outputCol" in key.name or "labelCol" in key.name or "inputCol" in key.name or "labelCol" in key.name or 'lazyAnnotator' in key.name or 'storageref' in key.name: continue

                # print("pipe['"+ component_key +"'].set"+ str( key.name[0].capitalize())+ key.name[1:]+"("+str(p_map[key])+")" + " | Info: " + str(key.doc)+ " currently Configured as : "+str(p_map[key]) )
                # print("Param Info: " + str(key.doc)+ " currently Configured as : "+str(p_map[key]) )

                if type(p_map[key]) == str:
                    s1 = "pipe['" + component_key + "'].set" + str(key.name[0].capitalize()) + key.name[
                                                                                               1:] + "('" + str(
                        p_map[key]) + "') "
                else:
                    s1 = "pipe['" + component_key + "'].set" + str(key.name[0].capitalize()) + key.name[1:] + "(" + str(
                        p_map[key]) + ") "

                s2 = " | Info: " + str(key.doc) + " | Currently set to : " + str(p_map[key])
                if len(s1) > max_len: max_len = len(s1)
                component_outputs.append((s1, s2))

            all_outputs.append((s, component_outputs))

        # make strings aligned
        form = "{:<" + str(max_len) + "}"
        for o in all_outputs:
            print(o[0])  # component name
            for o_parm in o[1]:
                if len(o_parm[0]) < max_len:
                    print(form.format(o_parm[0]) + o_parm[1])
                else:
                    print(o_parm[0] + o_parm[1])
    def print_exception_err(self,err):
        '''Print information about exception during converting or transforming dataframe'''
        import sys
        logger.exception('Exception occured')
        e = sys.exc_info()
        print("No accepted Data type or usable columns found or applying the NLU models failed. ")
        print("Make sure that the first column you pass to .predict() is the one that nlu should predict on OR rename the column you want to predict on to 'text'  ")
        print("On try to reset restart Jupyter session and run the setup script again, you might have used too much memory")
        print('Full Stacktrace was', e)
        print('Additional info:')
        exc_type, exc_obj, exc_tb = sys.exc_info()
        import os
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        print(exc_type, fname, exc_tb.tb_lineno)
        err = sys.exc_info()[1]
        print(str(err))
        print('Stuck? Contact us on Slack! https://join.slack.com/t/spark-nlp/shared_invite/zt-lutct9gm-kuUazcyFKhuGY3_0AMkxqA')











    def viz(self, text_to_viz:str, viz_type='', labels_to_viz=None,viz_colors={},return_html=False, write_to_streamlit=False, streamlit_key='NLU_streamlit'):
        """Visualize predictions of a Pipeline, using Spark-NLP-Display
        text_to_viz : String to viz
        viz_type    : Viz type, one of [ner,dep,resolution,relation,assert]
        labels_to_viz : Defines a subset of NER labels to viz i.e. ['PER'] , by default=[] which will display all labels. Applicable only for NER viz
        viz_colors  : Applicable for [ner, resolution, assert ] key = label, value=hex color, i.e. viz_colors={'TREATMENT':'#008080', 'problem':'#800080'}
        """
        from nlu.utils.environment.env_utils import install_and_import_package
        install_and_import_package('spark-nlp-display',import_name='sparknlp_display')
        if self.spark_transformer_pipe is None : self.fit()
        is_databricks_env = is_running_in_databricks()
        if return_html : is_databricks_env=True
        # self.configure_light_pipe_usage(1, force=True)
        from nlu.pipe.viz.vis_utils import VizUtils

        if viz_type == '' : viz_type  = VizUtils.infer_viz_type(self)
        # anno_res = self.spark_transformer_pipe.fullAnnotate(text_to_viz)[0]
        # anno_res = self.spark.createDataFrame(pd.DataFrame({'text':text_to_viz}))
        data, stranger_features, output_datatype = DataConversionUtils.to_spark_df(text_to_viz, self.spark, self.raw_text_column)
        anno_res = self.spark_transformer_pipe.transform(data)
        anno_res = anno_res.collect()[0]
        if self.has_licensed_components==False : HTML = VizUtils.viz_OS(anno_res, self, viz_type,viz_colors,labels_to_viz,is_databricks_env,write_to_streamlit, streamlit_key)
        else : HTML = VizUtils.viz_HC(anno_res, self, viz_type,viz_colors,labels_to_viz,is_databricks_env,write_to_streamlit)
        if return_html or is_databricks_env : return  HTML




    def viz_streamlit(self,
                      # Base Params
                      text:Union[str, List[str], pd.DataFrame, pd.Series] = "Angela Merkel from Germany and Donald Trump from America dont share many opinions",
                      model_selection:List[str] =[],
                      # SIMILARITY PARAMS
                      similarity_texts:Tuple[str,str]= ('I love NLU <3', 'I love Streamlit <3'),
                      # UI PARAMS
                      title:str = 'NLU ❤️ Streamlit - Prototype your NLP startup in 0 lines of code🚀',
                      sub_title:str = 'Play with over 1000+ scalable enterprise NLP models',
                      side_info:str = None,
                      visualizers:List[str] = ( "dependency_tree", "ner",  "similarity", "token_features", 'classification','manifold'),
                      show_models_info:bool = True,
                      show_model_select:bool = True,
                      show_viz_selection:bool = False,
                      show_logo:bool=True,
                      set_wide_layout_CSS:bool=True,
                      show_code_snippets:bool=False,
                      model_select_position:str = 'side',  # main or side
                      display_infos:bool=True,
                      key:str = "NLU_streamlit",
                      display_footer :bool =  True ,
                      num_similarity_cols:int=2,
                      ) -> None:
        """Display Viz in streamlit"""
        # try: from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        try: from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        except  ImportError : print("You need to install Streamlit to run this functionality.")
        StreamlitVizBlockHandler.viz_streamlit_dashboard(self,
                                          text,
                                          model_selection,
                                          similarity_texts,
                                          title,
                                          sub_title,
                                          side_info,
                                          visualizers,
                                          show_models_info,
                                          show_model_select,
                                          show_viz_selection,
                                          show_logo,
                                          set_wide_layout_CSS,
                                          show_code_snippets,
                                          model_select_position,
                                          display_infos,
                                          key,
                                          display_footer,
                                          num_similarity_cols
                                          )



    def viz_streamlit_token(
        self,
        text:str='NLU and Streamlit go together like peanutbutter and jelly',
        title: Optional[str] = "Token features",
        sub_title: Optional[str] ='Pick from `over 1000+ models` on the left and `view the generated features`',
        show_feature_select:bool =True,
        features:Optional[List[str]] = None,
        metadata: bool = True,
        output_level:str = 'token',
        positions:bool = False,
        set_wide_layout_CSS:bool=True,
        generate_code_sample:bool = False,
        key = "NLU_streamlit",
        show_model_select = True,
        model_select_position:str = 'side' , # main or side
        show_infos:bool = True,
        show_logo:bool = True,
        show_text_input:bool = True,


    ):
        try: from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        except  ImportError : print("You need to install Streamlit to run this functionality.")
        StreamlitVizBlockHandler.visualize_tokens_information(self, text, title, sub_title, show_feature_select, features, metadata, output_level, positions, set_wide_layout_CSS, generate_code_sample, key, show_model_select, model_select_position, show_infos, show_logo, show_text_input)


    def viz_streamlit_classes(
        self, # nlu pipe
        text:Union[str,list,pd.DataFrame, pd.Series, List[str]]=('I love NLU and Streamlit and sunny days!', 'I hate rainy daiys','CALL NOW AND WIN 1000$M'),
        output_level:Optional[str]='document',
        title: Optional[str] = "Text Classification",
        sub_title: Optional[str] = 'View predicted `classes` and `confidences` for `hundreds of text classifiers` in `over 200 languages`',
        metadata : bool = False,
        positions : bool = False,
        set_wide_layout_CSS:bool=True,
        generate_code_sample:bool = False,
        key:str = "NLU_streamlit",
        show_model_selector : bool = True ,
        model_select_position:str = 'side' ,
        show_infos:bool = True,
        show_logo:bool = True,
    )->None:
        try: from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        except  ImportError : print("You need to install Streamlit to run this functionality.")
        StreamlitVizBlockHandler.visualize_classes(self, text, output_level, title, sub_title, metadata, positions, set_wide_layout_CSS, generate_code_sample, key, show_model_selector, model_select_position, show_infos, show_logo)











    def viz_streamlit_dep_tree(
        self, #nlu pipe
        text:str = 'Billy likes to swim',
        title: Optional[str] = "Dependency Parse & Part-of-speech tags",
        sub_title: Optional[str] = 'POS tags define a `grammatical label` for `each token` and the `Dependency Tree` classifies `Relations between the tokens` ',
        set_wide_layout_CSS:bool=True,
        generate_code_sample:bool = False,
        key = "NLU_streamlit",
        show_infos:bool = True,
        show_logo:bool = True,
        show_text_input:bool = True,
    )->None:
        try: from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        except  ImportError : print("You need to install Streamlit to run this functionality.")
        StreamlitVizBlockHandler.visualize_dep_tree(self, text, title, sub_title, set_wide_layout_CSS, generate_code_sample, key, show_infos, show_logo, show_text_input, )

    def viz_streamlit_ner(
            self, # Nlu pipe
            text:str='Donald Trump from America and Angela Merkel from Germany do not share many views.',
            ner_tags: Optional[List[str]] = None,
            show_label_select: bool = True,
            show_table: bool = False,
            title: Optional[str] = "Named Entities",
            sub_title: Optional[str] = "Recognize various `Named Entities (NER)` in text entered and filter them. You can select from over `100 languages` in the dropdown.",
            colors: Dict[str, str] = {},
            show_color_selector: bool = False,
            set_wide_layout_CSS:bool=True,
            generate_code_sample:bool = False,
            key = "NLU_streamlit",
            model_select_position:str = 'side' , # main or side
            show_model_select = True,
            show_infos:bool = True,
            show_logo:bool = True,
            show_text_input:bool = True,

    ):
        try: from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        except ImportError : print("You need to install Streamlit to run this functionality.")
        StreamlitVizBlockHandler.visualize_ner(self, text, ner_tags, show_label_select, show_table, title, sub_title, colors, show_color_selector, set_wide_layout_CSS, generate_code_sample, key, model_select_position, show_model_select, show_infos, show_logo, show_text_input)


    def viz_streamlit_word_similarity(
            self, #nlu pipe
            texts: Union[Tuple[str, str], List[str]] = ("Donald Trump likes to party!", "Angela Merkel likes to party!"),
            threshold: float = 0.5,
            title: Optional[str] = "Vectors & Scalar Similarity & Vector Similarity & Embedding Visualizations  ",
            sub_tile :Optional[str]="Visualize a `word-wise similarity matrix` and calculate `similarity scores` for `2 texts` and every `word embedding` loaded",
            write_raw_pandas : bool = False ,
            display_embed_information:bool = True,
            similarity_matrix = True,
            show_algo_select : bool = True,
            dist_metrics:List[str]  =('cosine'),
            set_wide_layout_CSS:bool=True,
            generate_code_sample:bool = False,
            key:str = "NLU_streamlit",
            num_cols:int=2,
            display_scalar_similarities : bool = False ,
            display_similarity_summary:bool = False,
            model_select_position:str = 'side' ,
            show_infos:bool = True,
            show_logo:bool = True,



    ):
        try: from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        except ImportError : print("You need to install Streamlit to run this functionality.")
        StreamlitVizBlockHandler.display_word_similarity(self, texts, threshold, title, sub_tile, write_raw_pandas, display_embed_information, similarity_matrix, show_algo_select, dist_metrics, set_wide_layout_CSS, generate_code_sample, key, num_cols, display_scalar_similarities, display_similarity_summary, model_select_position, show_infos, show_logo, )


    def viz_streamlit_word_embed_manifold(self,
            default_texts: List[str] = ("Donald Trump likes to party!", "Angela Merkel likes to party!", 'Peter HATES TO PARTTY!!!! :('),
            title: Optional[str] = "Lower dimensional Manifold visualization for word embeddings",
            sub_title: Optional[str] = "Apply any of the 11 `Manifold` or `Matrix Decomposition` algorithms to reduce the dimensionality of `Word Embeddings` to `1-D`, `2-D` and `3-D` ",
            write_raw_pandas : bool = False ,
            default_algos_to_apply : List[str] = ('TSNE','PCA',),
            target_dimensions : List[int] = (1,2,3),
            show_algo_select : bool = True,
            show_embed_select : bool = True,
            show_color_select: bool = True,
            MAX_DISPLAY_NUM:int=100,
            display_embed_information:bool=True,
            set_wide_layout_CSS:bool=True,
            num_cols: int = 3,
            model_select_position:str = 'side', # side or main
            key:str = "NLU_streamlit",
            additional_classifiers_for_coloring:List[str]=['pos', 'sentiment.imdb'],
            generate_code_sample:bool = False,
            show_infos:bool = True,
            show_logo:bool = True,
            n_jobs: Optional[int] = 3, # False
        ):
        try: from nlu.pipe.viz.streamlit_viz.streamlit_dashboard_OS import StreamlitVizBlockHandler
        except ImportError : print("You need to install Streamlit to run this functionality.")
        StreamlitVizBlockHandler.viz_streamlit_word_embed_manifold(self,
        default_texts,
        title,
        sub_title,
        write_raw_pandas,
        default_algos_to_apply,
        target_dimensions,
        show_algo_select,
        show_embed_select,
        show_color_select,
        MAX_DISPLAY_NUM,
        display_embed_information,
        set_wide_layout_CSS,
        num_cols,
        model_select_position,
        key,
        additional_classifiers_for_coloring,
        generate_code_sample,
        show_infos,
        show_logo,
        n_jobs,)



import os
from os import path
import sys
PARENT_DIR = path.dirname(path.dirname(path.abspath(__file__)))
sys.path.append(PARENT_DIR)
conf_path = os.getcwd()
sys.path.append(conf_path)
sys.path.append('/home/loan/Documents/freelancework/jsl/nlu/4realnlugit/tmp/doc_generation')
# #nlu-1.0.2.tar.gz
import doc_generation.class_table_creator as class_table
import doc_generation.model_base_table_creator as model_table
import doc_generation.generate_bad_models_table  as bad_model_table



print("What do you want to do?\n"
      "1. Create Annotator Class Metadata Table(Could crash your machine, do manual runs)\n"
      "2. Generate Meta Models CSV. This can take 12-24 hours. Logs will be generated in runlogs.txt\nAssuming models will be default cached to '~/cache_pretrained'. If this not the case, configure the parameter manually (i.e. Windows) or your machine will crash\n"
      "3. Generate Bad Models CSV (CSV of all models for which no data could be generated in (2.)\n"
      "4. Exit"
      )


inp = int(input('Enter a number'))

if inp == 1 : class_table.generate()
if inp == 2:  model_table.generate(pretrained_model_cache_path = '~/cache_pretrained')
if inp == 3 : bad_model_table.generate()
if inp == 4:  exit()






import pandas as pd
p = '/home/loan/Documents/freelance/jsl/nlu/nlu4realgit/tmp/nlu_hc/all_models/HC_23_feb.csv'
df = pd.read_csv(p)

df['nlu_ref']= df.name.str.split("_")

df[['name','language']].sort_values(by=['name','language']).drop_duplicates().to_csv('sorted.csv',index=False)
df.to_csv('./resDf.csv')


SPARK_NLP_LICENSE =  "eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJleHAiOjE2NDIyNzM5ODksImlhdCI6MTYxMDczNzk4OSwidW5pcXVlX2lkIjoiYjQyZTc3ODgtNTc2NS0xMWViLTgyMDgtNzIyM2RkN2MyNzY0In0.vXrrwHdBssA7X1D7yfved7nKzvpxvKAOOKViNBS19S_DBoPRyBX1AwoQaisi-3Wp3MFnHZNKl6EVPLb3xt4UXLDjWs_5Nr6l32DAx1VuEZCAvtGqAJZeJsV7cgrRrf3Gh8WM2XutZRgsqQn21pNNGDcmxLH_-4LfPOqzrL5nNbZ2RXT_U3mD6umK38nD6gHaOCDn_zbZsum3SSZ0yUybA8OaCFTE8nPv-fdREBYHmM3mKYwmHguJxJcQTkSEfayMDnqx2G6k90ZOo4LcblC9wHPigF3WtsRpsRd1s2DEDu8r9rqmqK2Uxl1bHl38KgIQBhE0Z26qUTW8Hg031QUFOA"
AWS_ACCESS_KEY_ID =  "AKIASRWSDKBGNTFDKLUV"
AWS_SECRET_ACCESS_KEY =  "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"
OCR_LICENSE                 = 'eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJleHAiOjE2MzQyMjY4NDgsImlhdCI6MTYwMjYwNDQ0OCwidW5pcXVlX2lkIjoiNTQwYjU4ZDItMGQ2Yy0xMWViLWE0ZTAtNDJiYmNmMGU5Y2UyIn0.c-hgzBKdimyj7GtgDHQ3UjgNwSYWWdmKHlln3PFiu5b_b0wnodG2MJJ9rL1BXvBp2abW-6teQxA5j-7ZoB7Jl90xOMaoBtviUSvXdJMXQm7RN8y4zMFON2OnNDsfkt7kyyff5EI_7hD2M3pyDxtgr9DPaToXUR111o0MJq79_w_PG7GULNFZnNz73qXDw2fZpRG_aW0Akk6e8tJ2nkMP_YawUpLXSJHpGXMK6eBQ0qlBgk0y3ga__NqD062P6wyXXolrHQX40qXhA9vOnABUNz9PYzPMoGBdjmwrXMCbaSW8bX-dI5lLUOLQBcJSi4dQ7Jli_hhnHsdG4bIzZIVl9Q'
JSL_SECRET                  = '3.0.2-08d4231c2d0b95eaf306cf06e79d94f6567defe5'
# TODO TIME-SERIES-TOPIC-CHART -> TIME SERIES NER COUNT, LINE PLOT of HOW OFTEN EACH TOPIC OCCURS
# TODO VISUALIZE_STREAMLIT(HC=False) Visualized FULL UI with/withouth HC stuff  and code gen!
# TODO dropdown with ALL NLU REFS!
# TODO if RESOLUTION or RELATION PICEDK, make 2nd dropdown appear and make user pick NER model
# TODO, why not just make every viz applicable?
# TODO VECTOR SIMMILARITY
# TODO TOKEN ATTRIBUTES == JUST TABLE ---> BUT WITH TEXT AS HEADING!
import streamlit as st
import nlu
# TODO HYPERLINK OF MODELS TO MODELHUB!
# TODO MAKE MODEL SELECTION FOR ANY MODEL A MULTI SELECT!!
# TODO LATER FUNKY PLOTS
# Wind vector map! https://vega.github.io/vega-lite/examples/
# http://docs.bokeh.org/en/0.12.5/docs/reference/charts.html
# https://vega.github.io/vega-lite/examples/point_angle_windvector.html
nlu.auth(SPARK_NLP_LICENSE,AWS_ACCESS_KEY_ID,AWS_SECRET_ACCESS_KEY,JSL_SECRET)
# Cache Pipe! # https://docs.streamlit.io/en/stable/caching.html#the-hash-funcs-parameter

st.title('NLU ❤️ Streamlit - Prototype your NLP startup in 0 lines of code')

sample_data = {
'ner' : 'Donald Trump from America and Angela Merkel from Germany love John Snow Labs software!',
'med_ner' : 'Peter has no cancer!',
'dependency' : 'Billy likes to go to the mall',
'relation' : 'MRI demonstrated infarction in the upper brain stem , left cerebellum and  right basil ganglia',
'assertion' : '''The human KCNJ9 (Kir 3.3, GIRK3) is a member of the G-protein-activated inwardly rectifying potassium (GIRK) channel family.', 'Here we describe the genomicorganization of the KCNJ9 locus on chromosome 1q21-23 as a candidate gene forType II diabetes mellitus in the Pima Indian population.', 'The gene spansapproximately 7.6 kb and contains one noncoding and two coding exons separated byapproximately 2.2 and approximately 2.6 kb introns, respectively.', 'We identified14 single nucleotide polymorphisms (SNPs), including one that predicts aVal366Ala substitution, and an 8 base-pair', '(bp) insertion/deletion.', 'Ourexpression studies revealed the presence of the transcript in various humantissues including pancreas, and two major insulin-responsive tissues: fat andskeletal muscle.', 'The characterization of the KCNJ9 gene should facilitate furtherstudies on the function of the KCNJ9 protein and allow evaluation of thepotential role of the locus in Type II diabetes.''',
'resolution':'''The patient is a 5-month-old infant who presented initially on Monday with a cold, cough, and runny nose for 2 days. Mom states she had no fever. Her appetite was good but she was spitting up a lot. She had no difficulty breathing and her cough was described as dry and hacky. At that time, physical exam showed a right TM, which was red. Left TM was okay. She was fairly congested but looked happy and playful. She was started on Amoxil and Aldex and we told to recheck in 2 weeks to recheck her ear. Mom returned to clinic again today because she got much worse overnight. She was having difficulty breathing. She was much more congested and her appetite had decreased significantly today. She also spiked a temperature yesterday of 102.6 and always having trouble sleeping secondary to congestion.'''

}

def get_list_of_model_type(m_type='', licensed=True, get_all= False):
    nlu_refs_of_type = []
    model_universe = nlu.Spellbook.pretrained_models_references
    for lang, models in model_universe.items():
        for nlu_ref, nlp_ref in model_universe[lang].items():
            if m_type in nlu_ref or get_all: nlu_refs_of_type.append(nlu_ref)
    if licensed:
        model_universe = nlu.Spellbook.pretrained_healthcare_model_references
        for lang, models in model_universe.items():
            for nlu_ref, nlp_ref in model_universe[lang].items():
                if m_type in nlu_ref or get_all: nlu_refs_of_type.append(nlu_ref)
    return nlu_refs_of_type

@st.cache(allow_output_mutation=True)
def get_pipe(model='ner'): return nlu.load(model)
def viz_HTML_dep(text,model2viz='dep.typed',return_pandas=False):
    """Just VIZ the svg"""
    if not return_pandas: get_pipe(model2viz).viz(text, write_to_streamlit=True)
    else :
        df = get_pipe(model2viz).predict(text, output_level='token')
        st.write(df)
def viz_HTML_ner(text,model2viz,return_pandas=False):
    """Split the CSS and HTML and viz them seperatly or it wont render properly"""
    if not return_pandas: get_pipe(model2viz).viz(text, write_to_streamlit=True)
    else :
        df = get_pipe(model2viz).predict(text, output_level='chunk')
        st.write(df)
def viz_HTML_assertion(text,model2viz='en.med_ner.clinical en.assert',return_pandas=False):
    """Viz resolution"""
    if not return_pandas: get_pipe(model2viz).viz(text, write_to_streamlit=True)
    else :
        df = get_pipe(model2viz).predict(text, output_level='chunk')
        st.write(df)
def viz_HTML_resolution(text,model2viz='med_ner.jsl.wip.clinical en.resolve_chunk.cpt_clinical',return_pandas=False):
    """Viz resolution"""
    if not return_pandas: get_pipe(model2viz).viz(text, write_to_streamlit=True)
    # if not return_pandas:
    #     HTML = get_pipe(model2viz).viz(text, return_html=True)
    #     CSS,HTML = HTML.split('</style>')
    #     CSS = CSS + '</style>'
    #     HTML = f'<div> {HTML} '
    #     st.markdown(CSS, unsafe_allow_html=True)
    #     st.markdown(HTML, unsafe_allow_html=True)
    else :
        output_level = 'chunk' if 'chunk' in model2viz else 'sentence'
        df = get_pipe(model2viz).predict(text, output_level=output_level)
        st.write(df)
def viz_HTML_relation(text, model2viz='en.med_ner.jsl.wip.clinical.greedy en.relation',return_pandas=False):
    """Viz resolution"""
    # if not return_pandas:
    #     HTML = get_pipe(model2viz).viz(text, return_html=True)
    #     st.markdown(HTML, unsafe_allow_html=True)
    if not return_pandas: get_pipe(model2viz).viz(text, write_to_streamlit=True)
    else:
        df = get_pipe(model2viz).predict(text, output_level='relation')
        st.write(df)
def viz_HTML_any_model_raw(text, model2viz='emotion'):
    """Viz resolution"""
    df = get_pipe(model2viz).predict(text)
    st.write(df)

def make_base_UI():
    # Model Selection UI
    #
    ner_models         = get_list_of_model_type('ner')
    resolve_models     = get_list_of_model_type('resolve')
    assert_models      = get_list_of_model_type('assert')
    relation_models    = get_list_of_model_type('relation')
    max_text_display_len = 35
    text_for_model     = st.text_area("Enter text you want to visualize and process below", 'Billy loves Berlin')
    text_for_model_to_draw = text_for_model if len(text_for_model)< 100 else f"{text_for_model[:max_text_display_len]}..."
    actions  = ['Visualize a model','Run any NLU ref and view pandas','Run custom NLU ref and view pandas','T-SNE']
    selected_action = st.sidebar.selectbox('Select an action',(actions))

    def draw_custom_NLU_ref_UI():
        st.sidebar.write('Define any combination of nlu_references and run')
                                                                                                                                                                                                                                                                                                                                        custom_NLU2viz      = st.sidebar.text_input('Define a custom NLU reference to pass to nlu.load() and view RAW pandas','emotion sentiment')
        run = st.sidebar.button('Run model')
        if run :
            code =f"""#Running NLU command:\n import nlu \nimport streamlit as st \ndf=nlu.load('{custom_NLU2viz}').predict('{text_for_model_to_draw}')\nst.write(df)"""
            st.code(code, language='python')
            viz_HTML_any_model_raw(text_for_model,custom_NLU2viz)
    def draw_any_NLU_ref_UI():
        all_models          = get_list_of_model_type(get_all=True)
        any_model_2viz  = st.sidebar.selectbox('Pick any model to run',(all_models))
        run = st.sidebar.button('Run model')
        if run :
            code =f"""#Running NLU command:\n import nlu \nimport streamlit as st \nnlu.load('{any_model_2viz}').predict('{text_for_model_to_draw}')\nst.write(df)"""
            st.code(code, language='python')
            viz_HTML_any_model_raw(text_for_model,any_model_2viz)
    def draw_viz_ui():
        ner_model2viz      = st.sidebar.selectbox('Pick a Named Entity Recognizer (NER) model',(ner_models))
        resolve_model2viz  = st.sidebar.selectbox('Pick a Resolver model',(resolve_models))
        assert_model2viz   = st.sidebar.selectbox('Pick a Assertion model',(assert_models))
        relation_model2viz = st.sidebar.selectbox('Pick a Relation extractor model',(relation_models))
        models_types_to_run = ['Named Entity Recognizer (NER)','Dependency Tree and Part of Speech Tags','Resolution', 'Relation','Assertion']
        model_type2viz = st.sidebar.selectbox('Select a mode type to run and visualize. ',(models_types_to_run))
        st.sidebar.write('Assertion, Relation and Resolver models will receive input from currently selected NER model.')
        return_pandas = st.sidebar.checkbox('Return RAW pandas from NLU')
        if return_pandas: st.sidebar.write('Visualization disabled')
        run = st.sidebar.button('Run model')
        if run :
            if model_type2viz in ['ner','med_ner','Named Entity Recognizer (NER)']:
                if not return_pandas : code =f"""#Running NLU command:\nimport nlu \nnlu.load('{ner_model2viz}').viz('{text_for_model_to_draw}',write_to_streamlit=True)"""
                else :   code =f"""#Running NLU command:\nimport nlu\nimport streamlit as st \ndf=nlu.load('{ner_model2viz}').predict('{text_for_model_to_draw})\nst.write(df)"""

                st.code(code, language='python')
                viz_HTML_ner(text_for_model,ner_model2viz,return_pandas)
            if model_type2viz == 'Dependency Tree and Part of Speech Tags':
                code =f"""#Running NLU command:\nimport nlu \nnlu.load('dep.typed').viz('{text_for_model_to_draw},write_to_streamlit=True')"""
                st.code(code, language='python')
                viz_HTML_dep(text_for_model,'dep.typed',return_pandas) # there is only 1 dep
            # models below require NER input
            if model_type2viz == 'Resolution':
                code =f"""#Running NLU command:\nimport nlu \nnlu.load('{ner_model2viz + ' ' + resolve_model2viz}').viz('{text_for_model_to_draw},write_to_streamlit=True')"""
                st.code(code, language='python')
                viz_HTML_resolution(text_for_model,ner_model2viz + ' ' + resolve_model2viz,return_pandas)
            if model_type2viz == 'Assertion':
                code =f"""#Running NLU command:\nimport nlu \nnlu.load('{ner_model2viz + ' ' + assert_model2viz}').viz('{text_for_model_to_draw},write_to_streamlit=True')"""
                st.code(code, language='python')
                viz_HTML_assertion(text_for_model,ner_model2viz + ' ' +  assert_model2viz,return_pandas)
            if model_type2viz == 'Relation':
                code =f"""#Running NLU command:\nimport nlu \nnlu.load('{ner_model2viz + ' ' + relation_model2viz}').viz('{text_for_model_to_draw}, write_to_streamlit=True')"""
                st.code(code, language='python')
                viz_HTML_relation(text_for_model,ner_model2viz + ' ' +  relation_model2viz,return_pandas)
    def draw_T_SNE_UI():
        embed_models         = get_list_of_model_type('embed')
        #1. Get dataset?
        #2  Get label col for hue
        embed_model2viz      = st.sidebar.selectbox('Select an Embedding',(embed_models))
        pipe = get_pipe(embed_model2viz)
        def deduct_embed_col(): pass
        def deduct_embed_dimension(): pass
        def get_T_SNE_low_dim_data(start_dim, target_dim=2 ): pass

    if selected_action ==   'Visualize a model':                         draw_viz_ui()
    elif selected_action == 'Run any NLU ref and view pandas':   draw_any_NLU_ref_UI()
    elif selected_action == 'Run custom NLU ref and view pandas':draw_custom_NLU_ref_UI()
    elif selected_action == 'T-SNE':draw_T_SNE_UI()
make_base_UI()

def show_ui_ner(): pass
def hide_ui_ner(): pass
def show_ui_dep(): pass
def hide_ui_dep(): pass
def show_ui_resolution(): pass
def hide_ui_resolution(): pass
def show_ui_relation(): pass
def hide_ui_relation(): pass
# TODO, each model has UI with tunable params/ filters, etc.
# if model2viz == 'ner' : show_ui_ner()# TODO
# if model2viz == 'dep.typed' : show_ui_dep()# TODO
# if model2viz == 'resolution' : show_ui_resolution()# TODO
# if model2viz == 'relation' : show_ui_relation()# TODO
# if model2viz == 'dep.typed' : show_ui_dep()# TODO












# import pandas as pd
# import numpy as np
#
# chart_data = pd.DataFrame(
#     np.random.randn(20, 3),
#     columns=['a', 'b', 'c'])
#
# st.line_chart(chart_data)
#
# map_data = pd.DataFrame(
#     np.random.randn(1000, 2) / [50, 50] + [37.76, -122.4],
#     columns=['lat', 'lon'])
#
# st.map(map_data)
# options = st.multiselect(
#     'What are your favorite colors',
#     ['Green', 'Yellow', 'Red', 'Blue'],
#     ['Yellow', 'Red'])
#
# st.write('You selected:', options)







import unittest
from nlu import *

class TestLanguage(unittest.TestCase):

    def test_language_model(self):
        pipe = nlu.load('lang',verbose=True)
        data = ['NLU is an open-source text processing library for advanced natural language processing for the Python language.',
                'NLU est une bibliothèque de traitement de texte open source pour le traitement avancé du langage naturel pour les langages de programmation Python.',
                'NLU ist eine Open-Source Text verarbeitungs Software fuer fortgeschrittene natuerlich sprachliche Textverarbeitung in der Python Sprache '
                ]
        df = pipe.predict(data, output_level='sentence',drop_irrelevant_cols=False, metadata=True, )
        for c in df.columns: print(df[c])
        df = pipe.predict(data, output_level='document')
        for c in df.columns: print(df[c])


if __name__ == '__main__':
    unittest.main()







import unittest
from nlu import *

class TestDepTyped(unittest.TestCase):

    def test_dependency_typed_model(self):
        # This test takes too much ram on standard github actions machine
        df = nlu.load('dep.typed',verbose=True).predict('I love peanutbutter and jelly', output_level='sentence',drop_irrelevant_cols=False, metadata=True, )
        for c in df.columns: print(df[c])


        print("SENTENCE")
        df = nlu.load('dep.typed',verbose=True).predict('I love peanutbutter and jelly', output_level='sentence')
        for c in df.columns: print(df[c])

        print("TOKEN")
        df = nlu.load('dep.typed',verbose=True).predict('I love peanutbutter and jelly', output_level='token')
        for c in df.columns: print(df[c])


if __name__ == '__main__':
    unittest.main()




from sparknlp_jsl.annotator import ChunkEntityResolverModel,ChunkEntityResolverApproach

class ChunkResolver:
    @staticmethod
    def get_pretrained_model(name, language, bucket='clinical/models'):
        return ChunkEntityResolverModel.pretrained(name, language,bucket) \
            .setInputCols(["token","chunk_embeddings"]) \
            .setNeighbours(3) \
            .setOutputCol("chunk_resolution")

    @staticmethod
    def get_default_trainable_model():
        return ChunkEntityResolverApproach() \
        .setInputCols("token","chunk_embeddings") \
        .setOutputCol("chunk_resolution") \
        .setDistanceFunction("COSINE") \
            .setNeighbours(1) \
            .setLabelCol('label')



from nlu.pipe.pipe_components import SparkNLUComponent
class Resolver(SparkNLUComponent):
    def __init__(self, annotator_class='sentence_entity_resolver', language='en', component_type='resolution', get_default=True, model = None, nlp_ref ='', nlu_ref='',trainable=False, is_licensed=True, loaded_from_pretrained_pipe=False):

        if 'resolve' in nlu_ref and 'resolve_chunk' not in nlu_ref:
            annotator_class='sentence_entity_resolver'
        if 'resolve_chunk' in nlu_ref:
            annotator_class='chunk_entity_resolver'

        if model != None : self.model = model
        else :
            if annotator_class == 'sentence_entity_resolver':
                from nlu.components.resolutions.sentence_entity_resolver.sentence_resolver import SentenceResolver
                if trainable : self.model = SentenceResolver.get_default_trainable_model()
                elif get_default : self.model = SentenceResolver.get_default_model()
                else : self.model = SentenceResolver.get_pretrained_model(nlp_ref, language)
            elif annotator_class == 'chunk_entity_resolver':
                from nlu.components.resolutions.chunk_entity_resolver.chunk_resolver import ChunkResolver
                if trainable : self.model = ChunkResolver.get_default_trainable_model()
                elif get_default : self.model = ChunkResolver.get_default_model()
                else : self.model = ChunkResolver.get_pretrained_model(nlp_ref, language)

        SparkNLUComponent.__init__(self, annotator_class, component_type, nlu_ref, nlp_ref, language,loaded_from_pretrained_pipe , is_licensed)



from sparknlp.annotator import *

class ViveknSentiment:
    @staticmethod
    def get_default_model():
        return ViveknSentimentModel.pretrained() \
            .setInputCols(["sentence", "token"]) \
            .setOutputCol("sentiment")

    @staticmethod
    def get_pretrained_model(name, language):
        return ViveknSentimentModel.pretrained(name,language) \
            .setInputCols(["sentence", "token"]) \
            .setOutputCol("sentiment")


    @staticmethod
    def get_default_trainable_model():
        return ViveknSentimentApproach() \
            .setInputCols(["sentence", "token"]) \
            .setOutputCol("sentiment")\
            .setSentimentCol("sentiment_label")



from nlu import *

from dataclasses import dataclass
import glob
import os
import json
import sys
import logging
COMPONENT_INFO_FILE_NAME = 'component_infos.json'
logger = logging.getLogger('nlu')

class AllComponentsInfo:
    def __init__(self):
        ''' Initialize every NLU component info object and provide access to them'''

        self.all_components = {}
        self.classifiers = {}
        self.embeddings = {}
        self.normalizers  = {}
        self.pretrained_pipelines = {}
        self.selectors = {}
        self.spell_checkers = {}
        self.stemmers = {}
        self.tokenizers = {}
        self.utils = {}
        self.all_right_to_left_langs_with_pretrained_tokenizer = ['zh','ko','ja']
        self.all_pretrained_pipe_languages = ['en', 'nl','fr','de','it','no','pl','pt','ru','es','xx',]
        self.all_pretrained_model_languages = ['vi','mt','ta','af','cy','et','bh','am','da','fr','de','it','nb','no','nn','pl','pt','ru','es','af','ar','hy','eu','bn','br','bg','ca','cs','eo','fi','gl','el','ha','he','hi','hu','id','ga','ja','la','lv','mr','fa','ro','sk','sl','so','st','sw','sv','th','tr','uk','yo','zu','zh','xx','ur','ko',
                                               ]
        self.all_languages = set(self.all_pretrained_pipe_languages).union(set(self.all_pretrained_model_languages))
        self.all_classifier_classes =[]
        # this maps a requested token to a class
        self.all_nlu_actions = ['tokenize', 'pos', 'ner', 'embed', 'classify', 'sentiment', 'emotion', 'spell', 'dependency','dep','dep.untyped', 'match','sentence_detector', 'spell', 'stopwords'
                                    'labled_dependency','lemma', 'norm', 'select', 'pretrained_pipe','util', 'embed_sentence','embed_chunk','ngram']


        all_component_paths_regex = nlu.nlu_package_location + 'components/*/*/'
        all_component_paths = glob.glob(all_component_paths_regex)

        for path in all_component_paths :
            if '__py' in path : continue
            # logger.info('Loading info dict @ path'+ path)
            component = ComponentInfo.from_directory(path)
            self.all_components[component.name] = component
            if component.type == 'classifier' : self.classifiers[component.name] = component
            if component.type == 'embedding' : self.embeddings[component.name] = component
            if component.type == 'normalizer' : self.normalizers[component.name] = component
            if component.type == 'pretrained_pipeline' : self.pretrained_pipelines[component.name] = component
            if component.type == 'selector' : self.selectors[component.name] = component
            if component.type == 'spell_checker' : self.spell_checkers[component.name] = component
            if component.type == 'stemmer' : self.stemmers[component.name] = component
            if component.type == 'tokenizer' : self.tokenizers[component.name] = component
            if component.type == 'util' : self.utils[component.name] = component



    def list_all_components(self):
        print("--------------Avaiable Components in NLU :--------------")
        for name in self.all_components.keys(): print(name)

    def DEBUG_list_all_components(self):
        print("--------------Avaiable Components in NLU :--------------")
        for name in self.all_components.keys():
            print(name, " INPUT_F : ", self.all_components[name].inputs, " OUTPUT_F ", self.all_components[name].inputs," INPUT_N ", self.all_components[name].spark_output_column_names, "OUTPUT_N ", self.all_components[name].spark_output_column_names)

    def get_component_info_by_name(self,name): return self.all_components[name]
    def list_all_components_of_type(self,component_type='embeddings'): pass
    @staticmethod
    def list_all_components_of_language(component_lang='ger'): pass
    @staticmethod
    def list_all_components_of_languageand_type(component_lang='ger', component_type='embeddings'): pass
    @staticmethod
    def get_default_component_of_type():pass
    @staticmethod
    def list_avaiable_output_types():pass
    @staticmethod
    def get_all_component_info_obj():pass


@dataclass
class ComponentInfo:
    name: str
    description: str  # general annotator/model/component/pipeline info
    outputs: list  # this is which columns/output types this component is providing
    inputs: list  # this tells us which columns/input types the component is depending on
    type: str  # this tells us which kind of component this is
    output_level : str # document, sentence, token, chunk, input_dependent or model_dependent
    spark_input_column_names: list  # default expected name for input columns when forking with spark nlp annotators on spark DFs
    spark_output_column_names: list  # default expected name for output columns when forking with spark nlp annotators on spark DFs


    provider : str  # Who provides the implementation of this annotator, Spark-NLP for base. Would be
    license : str # open source or private
    computation_context : str # Will this component do its computation in Spark land (like all of Spark NLP annotators do) or does it require some other computation engine or library like Tensorflow, Numpy, HuggingFace, etc..
    output_context : str # Will this components final result
    trainable : bool

    @classmethod
    def from_directory(cls, component_info_dir):
        """Create ComponentInfo  class from the component_infos.json which is provided for every component
        @param component_info_dir:
            dataset_info_dir: `str` The directory containing the metadata file. This
            should be the root directory of a specific dataset version.
        """
        if not component_info_dir:
            raise ValueError("Calling DatasetInfo.from_directory() with undefined dataset_info_dir.")
        component_info_dir = component_info_dir.replace('//','/')
        with open(os.path.join(component_info_dir, COMPONENT_INFO_FILE_NAME), "r") as f:
            dataset_info_dict = json.load(f)

        try:
            return cls(**dataset_info_dict)# dataset_info_dict
        except :
            print (" Exception Occured! For Path",component_info_dir , " Json file most likely has missing  features. Todo nicer output error info", sys.exc_info()[0])
            raise



import nlu
from nlu.discovery import Discoverer
from nlu.pipe.utils.storage_ref_utils import StorageRefUtils
from typing import List, Tuple, Optional, Dict, Union
import streamlit as st
from nlu.utils.modelhub.modelhub_utils import ModelHubUtils
import numpy as np
import pandas as pd
from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
from nlu.pipe.viz.streamlit_viz.gen_streamlit_code import get_code_for_viz
from nlu.pipe.viz.streamlit_viz.styles import _set_block_container_style
import random
from nlu.pipe.viz.streamlit_viz.streamlit_viz_tracker import StreamlitVizTracker
class DepTreeStreamlitBlock():

    @staticmethod
    def visualize_dep_tree(
            pipe, #nlu pipe
            text:str = 'Billy likes to swim',
            title: Optional[str] = "Dependency Parse & Part-of-speech tags",
            sub_title: Optional[str] = 'POS tags define a `grammatical label` for `each token` and the `Dependency Tree` classifies `Relations between the tokens` ',
            set_wide_layout_CSS:bool=True,
            generate_code_sample:bool = False,
            key = "NLU_streamlit",
            show_infos:bool = True,
            show_logo:bool = True,
            show_text_input:bool = True,
    ):
        StreamlitVizTracker.footer_displayed=False
        if show_logo :StreamlitVizTracker.show_logo()
        if set_wide_layout_CSS : _set_block_container_style()
        if title:st.header(title)
        if show_text_input : text = st.text_area("Enter text you want to visualize dependency tree for ", text, key=key)
        if sub_title:st.subheader(sub_title)
        if generate_code_sample: st.code(get_code_for_viz('TREE',StreamlitUtilsOS.extract_name(pipe),text))
        pipe.viz(text,write_to_streamlit=True,viz_type='dep', streamlit_key=key)
        if show_infos :
            # VizUtilsStreamlitOS.display_infos()
            StreamlitVizTracker.display_model_info(pipe.nlu_ref, pipes = [pipe])
            StreamlitVizTracker.display_footer()







from sparknlp_jsl.annotator import  GenericClassifierModel, GenericClassifierApproach
from sparknlp_jsl.base import *
class GenericClassifier:
    @staticmethod
    def get_default_model():
        return GenericClassifierModel.pretrained() \
            .setInputCols("feature_vector") \
            .setOutputCol("generic_classification") \


    @staticmethod
    def get_pretrained_model(name, language):
        return GenericClassifierModel.pretrained(name,language) \
            .setInputCols("feature_vector") \
            .setOutputCol("generic_classification") \


    @staticmethod
    def get_default_trainable_model():
        return GenericClassifierApproach() \
            .setInputCols("feature_vector") \
            .setOutputCol("generic_classification") \
            .setLabelColumn("y") \
            .setEpochsNumber(2)






import unittest
from nlu import *

class TestDepUntyped(unittest.TestCase):

    def test_dependency_untyped_model(self):
        # This test takes too much ram on standard github actions machine
        return
        df = nlu.load('dep.untyped',verbose=True).predict('I love peanutbutter and jelly', output_level='document')
        for c in df.columns: print(df[c])


        df = nlu.load('dep.untyped',verbose=True).predict('I love peanutbutter and jelly', output_level='sentence')
        for c in df.columns: print(df[c])

        df = nlu.load('dep.untyped',verbose=True).predict('I love peanutbutter and jelly', output_level='token')
        for c in df.columns: print(df[c])






if __name__ == '__main__':
    unittest.main()




from sparknlp.annotator import *

class WordSegmenter:
    @staticmethod
    def get_default_model():
        return WordSegmenterModel()\
            .setInputCols(["document"]) \
            .setOutputCol("token")


    @staticmethod
    def get_pretrained_model(name, language):
        return WordSegmenterModel.pretrained(name,language) \
            .setInputCols(["document"]) \
            .setOutputCol("token")

    @staticmethod
    def get_default_model_for_lang(language):
        name = WordSegmenter.get_default_word_seg_for_lang(language)

        return WordSegmenterModel.pretrained(name,language) \
            .setInputCols(["document"]) \
            .setOutputCol("token")


    @staticmethod
    def get_default_word_seg_for_lang(language):
        import nlu
        # get default reference
        return nlu.Spellbook.pretrained_models_references[language][language + '.' + 'segment_words']